/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.container; 

import java.lang.Error;
import java.lang.RuntimeException;
import java.rmi.RemoteException;
import javax.ejb.EntityBean;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import javax.transaction.Status;
import javax.transaction.SystemException;
import javax.transaction.Synchronization;
import org.objectweb.jonas.common.Trace;


/**
 * This class implements the Synchronization interface for the Entity Beans
 */
public class EntitySynchroImpl extends SynchroImpl {
    
    EntityBean instance = null; // the bean instance
    JBeanEntity ejbobj = null; // interposition class
    boolean isContainerManaged = false;
    boolean ismarkedremoved = false;  // flag =true between a remove and the commit 

    /** 
     * Constructor on EntitySynchroImpl is called when the bean is created, that
     *  is when create() is invoked on Home interface.
     * EntitySynchroImpl object and Bean object are associated.
     * 
     */
    public EntitySynchroImpl(EntityBean bean, boolean containerManaged, JBeanEntity bullbean) {
	Trace.outln(trace, "EntitySynchroImpl.EntitySynchroImpl("+bean+" containerManaged= "+containerManaged+")");

	beanCoord = null;
	try {
	    currentCoord = tm.getTransaction();
	} catch (SystemException e) {
	    Trace.errln("EntitySynchroImpl: cannot get Transaction"+e);
	}

	instance = bean;
	isContainerManaged = containerManaged; // true if container managed persistence
	ejbobj = bullbean;
     }
   
    // -------------------------------------------------------------------
    // Synchronization implementation
    // -------------------------------------------------------------------

   /** 
     * This beforeCompletion method is called by the transaction 
     * manager prior to the start of the transaction completion process.
     * This method executes in the transaction context of the calling 
     * thread.
     * The Bean's state must be stored on the persistent storage before
     * the completion of the transaction.
    */
    public void beforeCompletion() {

	Trace.outln(trace, "EntitySynchroImpl.beforeCompletion()");

	// Persistence (always called)
       boolean isModified = ((JBeanEntity)ejbobj).isModifiedData(instance);
	try {
	    instance.ejbStore();
	    if (isContainerManaged && isModified) {
		ejbobj.storeData();
	    }
	} catch( RemoteException e) {
	    Trace.errln("EntitySynchroImpl.beforeCompletion() : " + e);
	    onException(e);
	} catch(RuntimeException e) {
	    Trace.errln("EntitySynchroImpl.beforeCompletion() : " + e);
	    onException(e);
 	} catch(Error e) {
	    Trace.errln("EntitySynchroImpl.beforeCompletion() : " + e);
	    onException(e);
	}
    }


    /** 
     * The afterCompletion method is called by the transaction 
     * manager after the transaction is committed or rolled back. 
     * This method executes without a transaction context.
     * Here we can update container tables used for handling multiple instances
     * of a bean.
     * If the bean has bean marked to be remove and the status of the transaction
     * is STATUS_COMMITTED we can remove the EJBObject in the table primarykey-EJBObject
     *
     * @param status The status of the transaction completion.
     */
    public void afterCompletion(int status) {
	Trace.outln(trace, "EntitySynchroImpl.afterCompletion(Status "+status+") ");

	// The main thing to do here is to release the instance
	try {
	    ejbobj.releaseInstance(beanCoord);
	} catch(Exception e) {
	    Trace.errln("EntitySynchroImpl.afterCompletion : releaseInstance fails:"+e);
	}

	// If this instance has been marked to be removed
	// during a transaction committed we can actually remove the EJBObject
	if (ismarkedremoved && (status == Status.STATUS_COMMITTED)) {
	    try {
		ejbobj.getHome().removeEJBObject(ejbobj.getPrimaryKey());
	    } catch (RemoteException e) {
		Trace.errln("EntitySynchroImpl.afterCompletion : removeEJBObject fails:"+e);
	    }
	    ejbobj.noMoreInUse(true);
	}
	ismarkedremoved = false;
    }


    // -------------------------------------------------------------------
    // Other methods
    // -------------------------------------------------------------------

    /**
     * A possible method for implementing Rules for container provider on Exceptions
     * cf spec 12.3.3 & 12.3.4
     *
     */
    public void onException (Throwable ue) {
	// TO BE DONE
	return;
    }

   /** 
    * This method considers the incoming bean request, and manages the 
    * synchrinization jobs.
    * On the first request we have to load the state of instance
    * May be used for Bean managed or Container managed persistence
    * We implements Option B for the three alternative commit options
    * described p 83 of the specs: 
    * "the container must synchronize the instance's state from the persistent
    * storage at the beginning of the next transaction"
    */
    public void considerRequest() throws RemoteException {
	Trace.outln(trace, "EntitySynchroImpl.considerRequest() ");
	
	// If this is not the first request for this Tx, return.
	// Note that if the request is called outside any Tx, it's like
	// if each request were in a separate Tx.
	if (beanCoord != null) {
	    return;
	}

	// Container managed persistence
	if (isContainerManaged) {
	    Trace.outln(trace, "Container managed persistence : call to loadData()");
	    ejbobj.loadData();
	}

	// Persistence (always called)
	try {
	    instance.ejbLoad();  // cf diagramm p86
	} catch( RemoteException e){
	    Trace.errln("EntitySynchroImpl.considerRequest() : " + e);
	    onException(e); 
	    throw e;
	} catch(RuntimeException e){
	    Trace.errln("EntitySynchroImpl.considerRequest() : " + e);
	    onException(e); 
	    throw new RemoteException("Get " +e.toString()+" thrown by ejbLoad()", e);
	} catch(Error e){
	    Trace.errln("EntitySynchroImpl.considerRequest() : " + e);
	    onException(e);
	    throw new RemoteException("Get " +e.toString()+" thrown by ejbLoad()", e);
	}
	
	// No transaction => no synchronization.
	try {
	    currentCoord = tm.getTransaction();
	} catch (SystemException e) {
	    Trace.errln("EntitySynchroImpl: cannot get Transaction:"+e);
	}
	if (currentCoord == null) {
	    Trace.outln(trace, "EntitySynchroImpl.considerRequest() currentCoord == null");
	    return;
	}
	
	// just to remember that next requests are not the first request.
	beanCoord = currentCoord;
	
	// register this object as a synchronization
	try {
	    currentCoord.registerSynchronization(this);
	    Trace.outln(trace, "EntitySynchroImpl.considerRequest() registerSynchronization("+this+")");
	} catch (Exception e) {
	    Trace.errln("EntitySynchroImpl.considerRequest() registerSynchronization(this) : "+e);
	    throw new RemoteException("Container cannot register synchronization Object", e);
	}
    }


    /**
     * Tests if the instance is marked to be removed
     *
     * @return false if the instance must be removed
     *
     */
    public boolean isMarkedRemoved() {
	return ismarkedremoved;
    }

    /**
     * Set the flag marked to be removed
     *
     * @param flg - true if a remove occurs on this instance
     *
     */
    public void setMarkedRemoved(boolean flg) {
	ismarkedremoved = flg;
    }

}
