/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.container;

import java.lang.reflect.Method;
import java.rmi.RemoteException;
import java.rmi.NoSuchObjectException;
import java.rmi.server.UnicastRemoteObject;
import java.util.Hashtable;
import javax.ejb.EJBHome;
import javax.ejb.EJBObject;
import javax.ejb.Handle;
import javax.ejb.RemoveException;
import javax.transaction.TransactionManager;
import org.objectweb.jonas.common.Trace;


/**
 * This class implements javax.ejb.EJBObject class.
 * This class is extended by all the interposition generated classes 
 *  corresponding to an EJB.
 * It mainly implements two operations <preinvoke> and <postinvoke>
 *  that are called before and after the business method of the Bean.
 */
public abstract class JBean extends UnicastRemoteObject implements EJBObject {

    private static final int trace = Trace.DB_17;

    /**
     * @serial
     */
    TransactionManager tm = Container.current;

    // Must be declared as "protected" because accessed by generated classes

    /**
     * @serial
     */
    protected JBeanHome home = null;

    /**
     * @serial
     */
    protected Handle handleEJBObject = null;

    /**
     * @serial
     */
    protected Object primarykey = null;

    /**
     * true if remote object no more accessible
     * @serial
     */
    protected boolean nosuchobject = false;
   
    /**
     * Constructor for a type of bean
     *
     * @param      EJBhome home
     * @exception  RemoteException
     *
     */
    JBean(JBeanHome home) throws RemoteException {
	Trace.outln(trace, this + ".JBean()");
	this.home = home;
    }

    // --------------------------------------------------------------------------
    // EJBObject implementation
    // --------------------------------------------------------------------------

    /**
     * Obtains the enterprise Bean's home interface. The home interface defines the enterprise
     *  Bean's create, finder, and remove operations.
     *
     * @return    A reference to the enterprise Bean's home interface. 
     * @exception RemoteException: Thrown when the method failed due to a system-level failure. 
     *
     */
    public EJBHome getEJBHome() throws RemoteException {
	if (nosuchobject)
	    throw new NoSuchObjectException("EJBObject already removed");
	return (EJBHome) home;  
    }


    /**
     * Obtains the primary key of the EJB object.
     *
     * @return    The EJB's object primary key.
     * @exception RemoteException: Thrown when the method failed due to a system-level failure. 
     *
     */
    abstract public Object getPrimaryKey() throws RemoteException;
  

    /**
     * Removes the EJB object.
     *
     * @exception RemoteException: Thrown when the method failed due to a system-level failure. 
     * @exception RemoveException: The enterprise Bean or the container does not allow 
     *            destruction of the object.
     *
     */
    abstract public void remove() throws RemoteException, RemoveException;


    /**
     * Obtains a handle for the EJB object. The handle can be used at later time to re-obtain
     *  a reference to the EHB object, possibly in a different Java Virtual Machine
     *
     * @return    A handle for the EJB object. 
     * @exception RemoteException: Thrown when the method failed due to a system-level failure. 
     *
     */
    abstract public Handle getHandle() throws RemoteException;


    /**
     * Tests if a given EJB is identical to the invoked EJB object.
     *
     * @param     EJBObject obj - An object to test for identity with the invoked object. 
     * @return    True if the given EJB object is identical to the invoked object, 
     *            false otherwise. 
     * @exception RemoteException: Thrown when the method failed due to a system-level failure. 
     *
     */
    abstract public boolean isIdentical(EJBObject obj) throws RemoteException;


    // --------------------------------------------------------------------------
    // Other methods
    // --------------------------------------------------------------------------




    /**
     * A possible method for 'cleaning' the state of the Bean after an unchecked exception
     * thrown by an enterprise Bean's method
     * (as described in EJB_Specifications & 12.3.3)
     * (Not yet implemented)
     *
     * @param ue the unchecked exception thrown by an enterprise Bean's method
     * @exception RemoteException: Thrown when the method failed due to a system-level failure. 
     *
     */
    public void onUncheckedException (Throwable ue) throws java.rmi.RemoteException {
	Trace.outln(trace, "JBean.onUncheckedException(Throwable "+ue.toString()+")");
	// TO BE DONE
	return;
    }


    /**
     * 
     * Set nosuchobject if the EJBObject must not be accessible
     */
    public void noMoreInUse(boolean flg){
	nosuchobject = flg;
    }
    
}
