/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.jdbc_xa;

import java.sql.Connection;
import java.sql.SQLException;
import javax.transaction.xa.*;
import org.objectweb.jonas.common.Trace;

/**
 * This is the implementation of the XAResource interface for our pseudo
 * JDBC-XA driver. This object is used to demarcate the transactions and
 * associate an XID to the SQL requests (start-end).
 * Actually, no real 2PC is done here because the std JDBC driver do not support
 * a java XA interface for it (only commit/rollback but no prepare!)
 * So we choose to return always OK at prepare. The goal is more to offer a
 * standard JDBC-XA interface to the EJB Container than to achieve a real
 * distributed 2PC. Anyway, when database vendors will offer real JDBC-XA
 * drivers, the EJBServer will be OK without any change.
 */
public class XAResourceImpl implements XAResource {

    private static final int trace = Trace.DB_10;

    private Connection actConn = null;
    private XAConnectionImpl xac = null;
    private int timeout;
    private Xid currXid = null;
    private boolean xa_started = false;
    private String rmid = null;

    // -----------------------------------------------------------------
    // Constructors
    // -----------------------------------------------------------------

    public XAResourceImpl(XAConnectionImpl xac, Connection actual, XADataSourceImpl ds) {

	Trace.outln(trace, "XAResourceImpl.XAResourceImpl(XAConnectionImpl xac, Connection actual, XADataSourceImpl ds)");

	this.xac = xac;
	this.actConn = actual;
	this.rmid = ds.getDataSourceName();
    }

    // -----------------------------------------------------------------
    // XAResource implementation
    // -----------------------------------------------------------------

    /**
     * Commit the global transaction specified by xid. 
     */
    public void commit(Xid xid, boolean onePhase) throws XAException {

	Trace.outln(trace, "XAResourceImpl.commit(Xid xid, boolean onePhase)");

	// Make sure that we do not mix transactions
	// Since start and end are not sent to database, we must check here
	// the xids.
	if (currXid != null && !currXid.equals(xid)) {
	    XAException ex = new XAException("Cannot mix transactions");
	    throw(ex);
	}
	if (xa_started) {
	    Trace.outln(trace, "### commit: XA START without XA END ###");
	}

	// Commit the transaction
	try {
	    actConn.commit();
	} catch (SQLException e) {
	    Trace.errln("Cannot commit transaction:"+e);
	    xac.notifyError(e);
	    throw(new XAException("Error on commit"));
	}
    }

    /**
     *  Ends the work performed on behalf of a transaction branch.
     */
    public void end(Xid xid, int flags) throws XAException {

	Trace.outln(trace, "XAResourceImpl.end(Xid xid, int flags)");

	// Check that we got a start on this Xid first.
	if (currXid == null || !currXid.equals(xid)) {
	    XAException ex = new XAException(XAException.XA_RBPROTO);
	    throw(ex);
	}
	xa_started = false;
    }

    /**
     * Tell the resource manager to forget about a heuristically 
     * completed transaction branch.
     */
    public void forget(Xid xid) throws XAException {

	// not implemented. XXX
	Trace.outln(trace, "XAResourceImpl.forget(Xid xid)");
	currXid = null;
    }

    /**
     * Obtain the current transaction timeout value set for this 
     * XAResource instance.
     */
    public int getTransactionTimeout() throws XAException {

	Trace.outln(trace, "XAResourceImpl.getTransactionTimeout()");
	return timeout;
    }

    /** Determine if the resource manager instance represented by the
     * target object is the same as the resource manager instance
     * represented by the parameter xares
     *
     * @param xares An XAResource object
     *
     * @return True if same RM instance, otherwise false.
     */
    public boolean isSameRM(XAResource xares) throws XAException {

	Trace.outln(trace, "XAResourceImpl.isSameRM(XAResource xares)");

	// In this pseudo-driver, we must return true only if
	// both objects refer to the same XAResource, and not
	// the same Resource Manager, because actually, we must
	// send commit/rollback on each XAResource involved in
	// the transaction.
	if (xares.equals(this)) {
	    return true;
	}
	return false;
    }


    /**
     * Ask the resource manager to prepare for a transaction commit 
     * of the transaction specified in xid.
     */
    public int prepare(Xid xid) throws XAException {

	Trace.outln(trace, "XAResourceImpl.prepare(Xid xid)");
	// No 2PC on standard JDBC drivers
	return XA_OK;
    }

    /**
     * Obtain a list of prepared transaction branches from a resource 
     * manager.
     */
    public Xid[] recover(int flag) throws XAException {

	Trace.outln(trace, "XAResourceImpl.recover(int flag)");
	// Not implemented
	return null;
    }
    
    /** 
     * Inform the resource manager to roll back work done on behalf 
     * of a transaction branch
     */
    public void rollback(Xid xid) throws XAException {

	Trace.outln(trace, "XAResourceImpl.rollback(Xid xid)");

	// Make sure that we do not mix transactions
	// Since start and end are not sent to database, we must check here
	// the xids.
	if (currXid != null && !currXid.equals(xid)) {
	    XAException ex = new XAException("Cannot mix transactions");
	    throw(ex);
	}
	if (xa_started) {
	    Trace.outln(trace, "### rollback: XA START without XA END ###");
	}

	// Make sure that we are not in AutoCommit mode
	try {
	    if (actConn.getAutoCommit() == true) {
		Trace.errln("Rollback called on XAResource with AutoCommit set");
		throw (new XAException(XAException.XA_HEURCOM));
	    }
	} catch (SQLException e) {
	    Trace.errln("Cannot getAutoCommit:"+e);
	    xac.notifyError(e);
	    throw(new XAException("Error on getAutoCommit"));
	}
	
	// Rollback the transaction
	try {
	    actConn.rollback();
	} catch (SQLException e) {
	    Trace.errln("Cannot rollback transaction:"+e);
	    xac.notifyError(e);
	    throw(new XAException("Error on rollback"));
	}
    }

    /** 
     * Set the current transaction timeout value for this XAResource
     * instance.
     */
    public boolean setTransactionTimeout(int seconds) throws XAException {

	Trace.outln(trace, "XAResourceImpl.setTransactionTimeout(int seconds)");
	timeout = seconds;
	return true;
    }

    /**
     * Start work on behalf of a transaction branch specified in xid
     */
    public void start(Xid xid, int flags) throws XAException {

	Trace.outln(trace, "XAResourceImpl.start(Xid xid, int flags)");
	if (xa_started) {
	    XAException ex = new XAException("mixed transactions");
	    throw ex;
	}
	currXid = xid;
	xa_started = true;
    }

    // ---------------------------------------------------------------------
    // Other methods
    // ---------------------------------------------------------------------

    /*
     * get the RMID
     */
    public String getRMID() {
	return rmid;
    }
}
