/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.jta;

import java.io.Serializable;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.Hashtable;
import java.util.Vector;
import javax.naming.Context;
import javax.naming.NamingException;
import javax.naming.Reference;
import javax.naming.Referenceable;
import javax.naming.StringRefAddr;
import javax.sql.ConnectionEvent;
import javax.sql.ConnectionEventListener;
import javax.sql.DataSource;
import javax.sql.XAConnection;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.xa.XAResource;
import org.objectweb.jonas.common.Trace;
import org.objectweb.jonas.jdbc_xa.XADataSourceImpl;

/**
 * This class represents a connection stored in the pool.
 * It may be in different states:
 * - available and reuseable for this user (open=false, tx=null)
 * - reusable for the same transaction, not already committed (open=false, tx!=null)
 * - still open, but no more involved in a Tx (open=true, tx=null)
 * - busy (open=true, tx!=null)
 * These 2 cases must be considered: close after commit, commit after close.
 * We cannot suppose what shema the user will use and we must handle both correctly.
 */
class PoolItem {
    public XAConnection xaConn;	// the pooled connection
    public String user;		// the user name
    public boolean open;	// true if connection is open
    public TransactionImpl tx;	// transaction the connection is involved with

    /*
     * constructor used by newConnection()
     */
    public PoolItem(XAConnection xac, String user) {
	this.xaConn = xac;
	this.user = user;
    }
}

class Pool {
    private static final int trace = Trace.DB_5;
    private Vector mypool = new Vector();
    
    /**
     * Look up a free XAConnection for this user and this transaction.
     * If found, set the item to "busy".
     *
     * @return the free PoolItem or null if no item available
     */
    public PoolItem openConnection(String user, TransactionImpl tx) {

	// Search for a connection that is not already open
	for (int i = 0; i < mypool.size(); i++) {
	    PoolItem item = (PoolItem) mypool.elementAt(i);
	    if (item.user == user && !item.open) {
		// No transaction = available for same user
		if (item.tx == null) {
		    Trace.outln(trace, "  Pool: got a free Connection");
		    item.tx = tx;
		    item.open = true;
		    return item;
		}
		// Ok too if same transaction
		if (tx != null && item.tx.equals(tx)) {
		    Trace.outln(trace, "  Pool: reuse a Connection for this Tx");
		    item.open = true;
		    return item;
		}
	    }
	}
	Trace.outln(trace, "  Pool: cannot find a reusable Connection");
	return null;
    }

    /**
     * Add a new XAConnection in the pool
     */
    public void newConnection(XAConnection xac, String user, TransactionImpl tx) {
	PoolItem item = new PoolItem(xac, user);
	item.open = true;
	item.tx = tx;
	mypool.addElement(item);
	Trace.outln(trace, "  Pool: add a Connection for this Tx");
    }

    /**
     * The transaction has committed (or rolled back). We can return its
     * connections to the pool of available connections.
     */
    public void freeConnections(Transaction tx) {
	boolean found = false;
	for (int i = 0; i < mypool.size(); i++) {
	    PoolItem item = (PoolItem) mypool.elementAt(i);
	    if (item.tx != null && item.tx.equals(tx)) {
		item.tx = null;
		found = true;
		Trace.outln(trace, "  Pool: free a Connection for this Tx");
	    }
	}
	if (!found) {
	    Trace.outln(trace, "  Pool: no connection found to free for this Tx");
	}
    }

    /**
     * Mark a specific Connection in the pool as closed.
     * If it is no longer associated to a Tx, we can free it.
     *
     * @return tx the transaction the connection is involved in
     */
    public Transaction closeConnection(XAConnection xac) {
	for (int i = 0; i < mypool.size(); i++) {
	    PoolItem item = (PoolItem) mypool.elementAt(i);
	    if (item.xaConn.equals(xac)) {
		// The connection will be available only if not associated
		// to a transaction. Else, it will be reusable only for the
		// same transaction.
		item.open = false;
		TransactionImpl tx = (TransactionImpl) item.tx;
		if (tx != null) {
		    Trace.outln(trace, "  Pool: keep the Connection for this tx");
		} else {
		    Trace.outln(trace, "  Pool: free the Connection");
		    item.tx = null;
		}
		return item.tx;
	    }
	}
	Trace.errln("  Pool: no connection found to free");
	return null;
    }

}

/**
 * JOnAS generic implementation of DataSource to manage an underlying
 * JDBC-XA driver (XADataSource, XAConnection, XAResource, ...)
 * This class acts as a factory for Connection objects.
 * Its main goal is to manage a pool of XAConnection objects, to avoid closing
 * the actual Connection on DBMS, and to deal with distributed transactions.
 * This DataSource should be registered in JNDI (-> impl. Referenceable)
 */
public class ConnectionManager implements DataSource, Referenceable, Serializable,
					  ConnectionEventListener {

    private static final int trace = Trace.DB_8;

    // The static table of all the ConnectionManager objects
    // Each ResourceManager has its own ConnectionManager (i.e. DataSource)
    private static Hashtable cmList = new Hashtable();

    /**
     * @serial
     */
    private Current current = Current.getCurrent(); // unique Current object
    /**
     * @serial
     */
    private XADataSourceImpl xads = null; // underlaying XA DataSource
    /**
     * @serial
     */
    private Pool pool = new Pool();

    // ----------------------------------------------------------
    // properties
    // ----------------------------------------------------------

    /**
     * @serial for JNDI
     */
    private String dataSourceName = null;
    /**
     * @serial url for database
     */
    private String url = null;
    /**
     * @serial JDBC driver Class
     */
    private String className = null;
    /**
     * @serial default user
     */
    private String userName = null;
    /**
     * @serial defalt passwd
     */
    private String password = null;

    // -----------------------------------------------------------------
    // Constructors
    // -----------------------------------------------------------------

    /**
     * Constructor for Factory
     */
    public ConnectionManager() {
	Trace.outln(trace, "ConnectionManager.ConnectionManager()");

	// Create the underlaying JDBC-XA driver
	// We do not use JNDI here. (should we do ?)
	xads = new XADataSourceImpl();
    }

    // -----------------------------------------------------------------
    // Properties
    // -----------------------------------------------------------------
    public String getDataSourceName() {
	return dataSourceName;
    }
    public void setDataSourceName(String s) {
	Trace.outln(trace, "ConnectionManager.setDataSourceName: "+s);
	dataSourceName = s;
	xads.setDataSourceName(s); // for getRMID!

	// Add it to the list
	cmList.put(s, this);
    }

    public String getUrl() {
	return url;
    }
    public void setUrl(String s) {
	Trace.outln(trace, "ConnectionManager.setUrl(String "+s+")");
	url = s;
	// Forward it to JDBC-XA driver
	xads.setUrl(s);
    }

    public String getClassName() {
	return className;
    }
    public void setClassName(String s) {
	Trace.outln(trace, "ConnectionManager.setClassName(String "+s+")");
	className = s;
	// Forward it to JDBC-XA driver
	xads.setClassName(s);
    }

    public String getUserName() {
	return userName;
    }
    public void setUserName(String s) {
	Trace.outln(trace, "ConnectionManager.setUserName(String "+s+")");
	userName = s;
	// Forward it to JDBC-XA driver
	xads.setUserName(s);
    }

    public String getPassword() {
	return password;
    }
    public void setPassword(String s) {
	Trace.outln(trace, "ConnectionManager.setPassword(String "+s+")");
	password = s;
	// Forward it to JDBC-XA driver
	xads.setPassword(s);
    }


    // -----------------------------------------------------------------
    // DataSource implementation
    // -----------------------------------------------------------------

    /**
     * Attempt to establish a database connection.
     *
     * @return a Connection to the database
     *
     * @exception java.sql.SQLException - if a database-access error occurs.
     */
    public Connection getConnection() throws SQLException {
	Trace.outln(trace, "ConnectionManager.getConnection()");
	return getConnection(userName, password);
    }

    /**
     * Attempt to establish a database connection.
     * The pool management must be "synchronized".
     *
     * @param user - the database user on whose behalf the Connection is being made
     * @param password - the user's password
     *
     * @return a Connection to the database
     *
     * @exception SQLException - if a database-access error occurs.
     */
    public synchronized Connection getConnection(String username, String password) throws SQLException {
	Trace.outln(trace, "ConnectionManager.getConnection(String "+username+",String "+password+")");

	XAConnection xac = null;
	Transaction tx = null;
	try {
	    tx = current.getTransaction();
	} catch (SystemException e) {
	    Trace.errln("ConnectionManager: getTransaction failed:"+e);
	}

	// Check if available XAConnection in the pool for this user
	PoolItem pi = pool.openConnection(username, (TransactionImpl) tx);
	if (pi == null) {
	    Trace.outln(trace, "ConnectionManager: No more XAConnection in the pool");
	    // Get a new XAConnection
	    xac = xads.getXAConnection(username, password);
	    // Put it in the pool as in-use
	    pool.newConnection(xac, username, (TransactionImpl) tx);
	    // Register me as a ConnectionEventListener
	    xac.addConnectionEventListener(this);
	} else {
	    // We found a reusable XAConnection
	    // We are still registered as ConnectionEventListener
	    Trace.outln(trace, "ConnectionManager: Reuse an XAConnection from the pool");
	    xac = pi.xaConn;
	}

	Connection ret = xac.getConnection();

	// Enlist XAResource if we are actually in a transaction
	if (tx != null) {
	    // We are inside a transaction: no autocommit!
	    ret.setAutoCommit(false);
	    try {
		tx.enlistResource(xac.getXAResource());
	    } catch (Exception e) {
		Trace.errln("Cannot enlist XAResource:"+e);
		// What shall we do here ? XXX
	    }
	} else {
	    // We are not in a transaction yet: we just make a pre-registration.
	    ret.setAutoCommit(true);	// in case we do not start a Tx
	    current.addConnection(xac);
	}

	// return a Connection object
	return ret;
    }

    /**
     * Get the log writer for this data source. 
     * 
     * The log writer is a character output stream to which all logging and 
     * tracing messages for this data source object instance will be printed. 
     * This includes messages printed by the methods of this object, messages 
     * printed by methods of other objects manufactured by this object, and so on.
     * Messages printed to a data source specificlog writer are not printed to the 
     * log writer associated with the Drivermanager class. When a data
     * source object is created the log writer is initially null, in other words, 
     * logging is disabled.
     *
     * @return the log writer for this data source, null if disabled
     *
     * @exception SQLException - if a database-access error occurs.
     */
    public java.io.PrintWriter getLogWriter() throws SQLException {
	Trace.outln(trace, "ConnectionManager.getLogWriter()");
	return xads.getLogWriter();
    }


    /**
     * Set the log writer for this data source. 
     * 
     * The log writer is a character output stream to which all logging and 
     * tracing messages for this data source object instance will be printed. 
     * This includes messages printed by the methods of this object, messages printed
     * by methods of other objects manufactured by this object, and so on. 
     * Messages printed to a data source specific log writer are not printed to 
     * the log writer associated with the Drivermanager class. When a data
     * source object is created the log writer is initially null, in other words, 
     * logging is disabled.
     * 
     * @param out - the new log writer; to disable, set to null
     *
     * @exception SQLException - if a database-access error occurs.
     */
    public void setLogWriter(java.io.PrintWriter out) throws SQLException {
	Trace.outln(trace, "ConnectionManager.setLogWriter()");
	xads.setLogWriter(out);
    }


    /**
     * Sets the maximum time in seconds that this data source will wait while 
     * attempting to connect to a database.
     * A value of zero specifies that the timeout is the default system timeout 
     * if there is one; 
     * otherwise it specifies that there is no timeout. 
     * When a data source object is created the login timeout is initially zero.
     *
     * @param seconds - the data source login time limit
     *
     * @exception SQLException - if a database access error occurs.
     */
    public void setLoginTimeout(int seconds) throws SQLException {
	Trace.outln(trace, "ConnectionManager.setLoginTimeout(int "+seconds+")");
	xads.setLoginTimeout(seconds);
    }

    /**
     * Gets the maximum time in seconds that this data source can wait while attempting 
     * to connect to a database.
     * A value of zero means that the timeout is the default system timeout if there is one; 
     * otherwise it means that there is no timeout. When a data source object is created 
     * the login timeout is initially zero.
     *
     * @return the data source login time limit
     *
     * @exception SQLException - if a database access error occurs.
     */
    public int getLoginTimeout() throws SQLException {
	Trace.outln(trace, "ConnectionManager.getLoginTimeout()");
	return xads.getLoginTimeout();
    }

    // -----------------------------------------------------------------
    // ConnectionEventListener implementation
    // -----------------------------------------------------------------

    /**
     * Invoked when the application calls close() on its representation of the connection.
     *
     * @param event - an event object describing the source of the event
     */
    public void connectionClosed(ConnectionEvent event) {
	Trace.outln(trace, "ConnectionManager.connectionClosed(ConnectionEvent event)");

	// First of all, always mark the Connection as closed.
	XAConnection xac = (XAConnection) event.getSource();
	TransactionImpl tx = (TransactionImpl) pool.closeConnection(xac);
	// tx should be the same than getTransaction() ? XXX

	// delist Resource if in transaction
	// We must keep the connection till commit or rollback
	if (tx != null) {
	    try {
		tx.delistResource(xac.getXAResource(), XAResource.TMSUCCESS);
	    } catch (Exception e) {
		Trace.errln("Exception in connectionClosed:"+e);
	    }
	} else {
	    current.removeConnection(xac);
	}
    }


    /**
     * Invoked when a fatal connection error occurs, just before an SQLException is 
     * thrown to the application.
     *
     * @param event - an event object describing the source of the event
     */
    public void connectionErrorOccurred(ConnectionEvent event) {
	Trace.outln(trace, "ConnectionManager.connectionErrorOccurred(ConnectionEvent event)");

	// First of all, always mark the Connection as closed.
	XAConnection xac = (XAConnection) event.getSource();
	TransactionImpl tx = (TransactionImpl) pool.closeConnection(xac);

	// delist Resource if in transaction
	if (tx != null) {
	    try {
		tx.delistResource(xac.getXAResource(), XAResource.TMFAIL);
	    } catch (Exception e) {
		Trace.errln("Exception in connectionErrorOccurred:"+e);
	    }
	} else {
	    // Not sure we must do this...
	    current.removeConnection(xac);
	}
    }

    // -----------------------------------------------------------------
    // Other methods
    // -----------------------------------------------------------------

    /**
     * Forget all about this Transaction (after the commit or rollback)
     */
    public void cleanTransaction(Transaction tx) {
	Trace.outln(trace, "ConnectionManager.cleanTransaction(Transaction tx)");

	pool.freeConnections(tx);
    }

    // -----------------------------------------------------------------
    // Referenceable implementation
    // -----------------------------------------------------------------

    /**
     * Retrieves the Reference of this object. Used at binding time by JNDI
     * to build a reference on this object.
     * 
     * @return  The non-null Reference of this object. 
     * @exception  NamingException  If a naming exception was encountered while retrieving the reference.  
     */
    public Reference getReference() throws NamingException {

	Reference ref = new Reference (this.getClass().getName(), 
				       "org.objectweb.jonas.jta.DataSourceFactory",
				       null);
	// Actually, only the datasource.name is used! (See DataSourceFactory.java)
	ref.add(new StringRefAddr("datasource.name", getDataSourceName()));
	ref.add(new StringRefAddr("datasource.url", getUrl()));
	ref.add(new StringRefAddr("datasource.classname", getClassName()));
	ref.add(new StringRefAddr("datasource.username", getUserName()));
	ref.add(new StringRefAddr("datasource.password", getPassword()));
	return ref;
    }

   /**
    * get the ConnectionManager matching the DataSource name
    */
   public static ConnectionManager getConnectionManager(String dsname) {
       
       ConnectionManager cm = (ConnectionManager) cmList.get(dsname);
       return cm;
   }

}
