/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.jta;

import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.Vector;
import javax.sql.DataSource;
import javax.transaction.*;
import javax.transaction.xa.XAException;
import javax.transaction.xa.XAResource;
import javax.transaction.xa.Xid;
import org.objectweb.jonas.common.Trace;
import org.objectweb.jonas.jtm.HeuristicCommit;
import org.objectweb.jonas.jtm.Resource;

/**
 * Log associated to this transaction coordinator
 */
class Log {

    Vector resourceLogged = new Vector();
    int decision;
    static final int DECISION_TO_COMMIT = 1;
    static final int DECISION_TO_ROLLBACK = 2;

    public void addResource(XAResource res) {
	resourceLogged.addElement(res);
    }

    public void flushLog(int decide) {
	decision = decide;

	// serialize log on disk
	// XXX
    }    

    public void forgetLog() {

	// remove file on disk
	// XXX
    }
}

/**
 * This object is the local coordinator. It may be registered as
 * sub-coordinator in case of distributed transaction, so it must
 * be callable remotely and implement Resource
 */
public class SubCoordinator extends UnicastRemoteObject implements Resource {

    private static final int trace = Trace.DB_5;

    // ------------------------------------------------------------------
    // Object state
    // ------------------------------------------------------------------

    /**
     * @serial
     */
    private TransactionImpl tx = null;
    /**
     * @serial
     */
    private Vector synchroList = new Vector();		// Synchronization
    /**
     * @serial
     */
    private Vector resourceList = new Vector();		// XAResource
    /**
     * @serial
     */
    private Xid xid = null;
    /**
     * @serial
     */
    private Log log = null;
    /**
     * @serial
     */
    private int status = 0;	// javax.transaction.Status
    /**
     * @serial
     */
    private boolean beforeCompletionDone = false;

    // ------------------------------------------------------------------
    // Constructors
    // ------------------------------------------------------------------

    /**
     * constructor used by TransactionImpl
     */
    SubCoordinator(TransactionImpl tx, Xid xid) throws RemoteException {
	Trace.outln(trace, "SubCoordinator.SubCoordinator(TransactionImpl tx, Xid xid)");
	this.tx = tx;
	this.xid = xid;
	status = Status.STATUS_ACTIVE;
    }

    // ------------------------------------------------------------------
    // Resource interface
    // ------------------------------------------------------------------

    /** 
     * phase 1 of the 2PC.
     *
     * @return int vote commit, rollback, or readonly.
     */
    public int prepare() throws RemoteException {

	Trace.outln(trace, "SubCoordinator.prepare()");


	switch (status) {
	case Status.STATUS_MARKED_ROLLBACK:
	    doBeforeCompletion(false);
	    doRollback();
	    return Resource.VOTE_ROLLBACK;
	case Status.STATUS_COMMITTED:
	    return Resource.VOTE_COMMIT;
	default:
	    doBeforeCompletion(true);
	    break;
	}

	int ret = doPrepare();
	if (ret == Resource.VOTE_READONLY) {
	    // Transaction completed for this Resource
	    doAfterCompletion();
	}
	return ret;
    }

    /** 
     * rollback transaction
     */
    public void rollback() throws RemoteException {

	Trace.outln(trace, "SubCoordinator.rollback()");

	switch (status) {
	case Status.STATUS_ACTIVE:
	case Status.STATUS_MARKED_ROLLBACK:
	    break;
	case Status.STATUS_ROLLEDBACK:
	    return;
	default:
	    Trace.errln("rollback: bad status");
	    return;
	}

	doBeforeCompletion(false);
	doRollback();
    }

    /** 
     * phase 2 of the 2PC.
     */
    public void commit() throws RemoteException {

	Trace.outln(trace, "SubCoordinator.commit()");

	switch (status) {
	case Status.STATUS_PREPARED:
	    break;
	default:
	    Trace.errln("commit: bad status");
	    return;
	}

	doCommit();
    }


    /** 
     * commit 1 phase. Called either from JTM (distributed transaction) or
     * from Transaction.commit (local transaction).
     */
    public void commit_one_phase() throws RemoteException {

	Trace.outln(trace, "SubCoordinator.commit_one_phase()");
	

	switch (status) {
	case Status.STATUS_ROLLEDBACK:
	    throw new TransactionRolledbackException();
	case Status.STATUS_MARKED_ROLLBACK:
	    doBeforeCompletion(false);
	    doRollback();
	    throw new TransactionRolledbackException();
	case Status.STATUS_COMMITTED:
	    return;
	default:
	    doBeforeCompletion(true);
	    break;
	}

	// only 1 Resource => 1 phase commit
	if (resourceList.size() == 1) {
	    doOnePhaseCommit();
	    return;
	}

	// 2 phase commit
	int vote = doPrepare();

	switch (vote) {
	case Resource.VOTE_COMMIT:
	    doCommit();
	    break;
	case Resource.VOTE_READONLY:
	    doAfterCompletion();
	    break;
	case Resource.VOTE_ROLLBACK:
	    doRollback();
	    break;
	}
    }

    /** 
     * forget heuristics about this transaction.
     */
    public void forget() throws RemoteException {

	Trace.outln(trace, "SubCoordinator.forget()");
	doForget();
    }

    // ------------------------------------------------------------------
    // Other public methods (interface exposed to Transaction object)
    // ------------------------------------------------------------------

    /**
     * add a Synchronization to the list
     *
     * @param synchro The javax.transaction.Synchronization object for the 
     *    transaction associated with the target object
     *
     * @exception RollbackException Thrown to indicate that
     *    the transaction has been marked for rollback only.
     *
     * @exception IllegalStateException Thrown if the transaction in the 
     *    target object is in prepared state or the transaction is inactive.
     *
     */
    public void addSynchronization(Synchronization synchro) 
	throws RollbackException, IllegalStateException {

	Trace.outln(trace, "SubCoordinator.addSynchronization(Synchronization synchro)");

	//check status: should be ACTIVE.
	switch(status) {
	case Status.STATUS_MARKED_ROLLBACK:
	case Status.STATUS_ROLLEDBACK:
	    throw new RollbackException();
	case Status.STATUS_ACTIVE:
	    break;
	default:
	    throw new IllegalStateException("addSynchronization: Bad status");
	}

	// Add synchro to the list of local synchros
	synchroList.addElement(synchro);
    }

    /**
     * add a XAResource to the list
     *
     * @param the XAResource to register
     *
     * @return true if this datasource was already known
     *
     * @exception RollbackException Thrown to indicate that
     *    the transaction has been marked for rollback only.
     *
     * @exception IllegalStateException Thrown if the transaction in the 
     *    target object is in prepared state or the transaction is inactive.
     */
    public boolean addResource(XAResource xares) 
	throws RollbackException, IllegalStateException {

	Trace.outln(trace, "SubCoordinator.addResource(XAResource xares) ");

	//check status: should be ACTIVE.
	switch(status) {
	case Status.STATUS_MARKED_ROLLBACK:
	case Status.STATUS_ROLLEDBACK:
	    throw new RollbackException();
	case Status.STATUS_ACTIVE:
	    break;
	default:
	    throw new IllegalStateException("addResource: Bad status");
	}

	// Check if this DataSource is already known
	// -> we must register only ONE resource per DataSource.
	boolean found = false;
	for (int i = 0; i < resourceList.size(); i++) {
	    XAResource res = (XAResource) resourceList.elementAt(i);
	    try {
		if (res.isSameRM(xares)) {
		    found = true;
		    break;
		}
	    } catch (XAException e) {
		Trace.errln("Exception on isSameRM:"+e);
	    }
	}

	// add this XAResource to the list
	if (found == false) {
	    Trace.outln(trace, "SubCoordinator.addResource: new datasource");
	    resourceList.addElement(xares);
	}
	return found;
    }

    /**
     * return the status of this transaction
     */
    public int getStatus() {
	Trace.outln(trace, "SubCoordinator.getStatus()");
	return status;
    }

    /**
     * set the transaction "rollback only"
     */
    public void setRollbackOnly() {
	Trace.outln(trace, "SubCoordinator.setRollbackOnly()");
	switch (status) {
	case Status.STATUS_ACTIVE:
	case Status.STATUS_UNKNOWN:
	case Status.STATUS_PREPARING:
	    status = Status.STATUS_MARKED_ROLLBACK;
	    break;
	case Status.STATUS_MARKED_ROLLBACK:
	case Status.STATUS_PREPARED:
	case Status.STATUS_COMMITTED:
	case Status.STATUS_ROLLEDBACK:
	case Status.STATUS_NO_TRANSACTION:
	case Status.STATUS_COMMITTING:
	case Status.STATUS_ROLLING_BACK:
	    Trace.errln("Cannot set transaction as rollback only");
	    break;
	}
    }
	
    // ------------------------------------------------------------------
    // private methods
    // ------------------------------------------------------------------

    /**
     * forget Transaction
     */
    private void doForget() throws RemoteException {

	Trace.outln(trace, "SubCoordinator.doForget()");

	boolean exception = false;

	for (int i = 0; i < resourceList.size(); i++) {
	    // Tell this Resource to forget the transaction
	    // Remove from the list ? (XXX)
	    XAResource xar = (XAResource) resourceList.elementAt(i);
	    try {
		xar.forget(xid);
	    } catch (XAException e) {
		exception = true;
	    }
	}
	if (exception) {
	    throw new RemoteException("XAException on forget");
	}

    }

    /**
     * 2PC phase 1 (prepare)
     * Basically, send prepare on each XAResource and
     * collect the results.
     */
    private int doPrepare() throws RemoteException {

	Trace.outln(trace, "SubCoordinator.doPrepare()");

	int ret = VOTE_READONLY;
	int errors = 0;

	// No resource
	if (resourceList.size() == 0) {
	    status = Status.STATUS_COMMITTED;
	    return ret;
	}

	// Creates a log for that transaction, where we will add all the
	// resources that replied VOTE_COMMIT to prepare.
	// Do not flush the log on disk before decision to commit.
	log = new Log();

	// Sends prepare to each resource while no error
	// In case of prepare on sub-coord. we may have only 1 resource.
	status = Status.STATUS_PREPARING;
	for (int i = 0; i < resourceList.size(); i++) {
	    XAResource res = (XAResource) resourceList.elementAt(i);
	    if (errors > 0) {
		try {
		    res.rollback(xid);
		} catch (XAException e) {
		    // Nothing to do ?
		}
	    } else {
		try {
		    switch (res.prepare(xid)) {
		    case XAResource.XA_OK:
			log.addResource(res);
			ret = VOTE_COMMIT;
			break;
		    case XAResource.XA_RDONLY:
			break;
		    }
		} catch (XAException e) {
		    ret = VOTE_ROLLBACK;
		    errors++;
		}
	    }
	}

	// Update the status, depending on vote result
	// If all resources returned READ_ONLY, we can forget the transaction
	switch (ret) {
	case VOTE_READONLY:
	    status = Status.STATUS_COMMITTED;
	    break;
	case VOTE_COMMIT:
	    status = Status.STATUS_PREPARED;
	    break;
	case VOTE_ROLLBACK:
	    status = Status.STATUS_ROLLEDBACK;
	    break;
	}

	// return the global vote
	return ret;
    }

    /**
     * 2PC - phase 2 (commit)
     * See JTM for heuristics management XXX
     */
    private int doCommit() throws RemoteException {

	Trace.outln(trace, "SubCoordinator.doCommit()");

	// First check that a log is initialized
	if (log == null) {
	    Trace.errln("doCommit: no log");
	    return -1;
	}

	// Status Transaction = committing
	status = Status.STATUS_COMMITTING;

	// Send commit to each resource prepared
	int errors = 0;
	int commitnb = 0;
	int heuristicnb = 0;
	for (int i = 0; i < log.resourceLogged.size(); i++) {
	    XAResource res = (XAResource) log.resourceLogged.elementAt(i);

	    // aborts transaction if error and no resource committed yet
	    if (commitnb == 0 && errors > 0) {
		try {
		    res.rollback(xid);
		} catch (XAException e) {
		    commitnb++;
		    heuristicnb++;
		}
		continue;
	    }

	    // commit resource
	    try {
		res.commit(xid, false);
	    } catch (XAException e) {
		errors++;
		if (commitnb > 0)
		    heuristicnb++;
	    }
	}

	if (errors == 0) {
	    // Everything's fine.
	    status = Status.STATUS_COMMITTED;
	    doAfterCompletion();
	    log.forgetLog();
	    return 0;
	}

	if (heuristicnb == 0) {
	    // Transaction has been eventually rolled back
	    status = Status.STATUS_ROLLEDBACK;
	    doAfterCompletion();
	    log.forgetLog();
	    throw new TransactionRolledbackException();
	}

	// Log heuristics if errors
	// XXX
	status = Status.STATUS_UNKNOWN;
	doAfterCompletion();
	return -1;
    }

    /**
     * 1PC (commit one phase)
     */
    private void doOnePhaseCommit() throws RemoteException {

	Trace.outln(trace, "SubCoordinator.doOnePhaseCommit()");

	// Only 1 resource: commit with onePhase=true
	status = Status.STATUS_COMMITTING;
	XAResource res = (XAResource) resourceList.elementAt(0);
	try {
	    res.commit(xid, true);
	    status = Status.STATUS_COMMITTED;
	} catch (XAException e) {
	    status = Status.STATUS_UNKNOWN;
	    throw new RemoteException("XAException:"+e);
	}

	doAfterCompletion();
    }

    /**
     * Rollback every resource involved
     */
    private void doRollback() throws RemoteException {

	Trace.outln(trace, "SubCoordinator.doRollback()");

	status = Status.STATUS_ROLLEDBACK;
	boolean hcomm = false;

	// rollback each resource
	for (int i = 0; i < resourceList.size(); i++) {
	    XAResource res = (XAResource) resourceList.elementAt(i);
	    try {
		res.rollback(xid);
	    } catch (XAException e) {
		switch (e.errorCode) {
		case e.XA_HEURCOM:
		    hcomm = true;
		default:
		    throw new RemoteException("rollback: Unexpected XAException:"+e.errorCode);
		}
	    }
	}

	doAfterCompletion();

	// raise exception if error on a resource
	if (hcomm) {
	    throw new HeuristicCommit();
	}
    }

    /**
     * before completion
     *
     * @param boolean true if completion ok, false if rollback
     */
    private void doBeforeCompletion(boolean committing) {

	if (beforeCompletionDone)
	    return;

	Trace.outln(trace, "SubCoordinator.doBeforeCompletion()");

	// Unset the timer for this transaction, if any
	tx.unsetTimer();

	// For each synchro, send beforeCompletion (not if rollback)
	if (committing && synchroList.size() > 0) {
	    // We must be in the correct transaction context
	    // See JTA spec. page 13 (3.3.2)
	    // because at beforeCompletion, the bean will write its cache

	    // Check the trivial case where we already have the correct tx context
	    Transaction mytx = null;
	    boolean suspended = false;
	    boolean resumed = false;
	    try {
		mytx = Current.getCurrent().getTransaction();
	    } catch (SystemException e) {
		Trace.outln(trace, "Cannot get transaction:"+e);
	    }

	    // Suspend if another tx context
	    if (mytx != null && mytx.equals(tx) == false) {
		try {
		    Current.getCurrent().suspend();
		    suspended = true;
		} catch (SystemException e) {
		    Trace.outln(trace, "Cannot suspend transaction:"+e);
		}
	    }

	    // Resume the good tx context
	    if (mytx == null || suspended) {
		try {
		    Current.getCurrent().resume(tx);
		    resumed = true;
		} catch (SystemException e) {
		    Trace.outln(trace, "Cannot resume transaction:"+e);
		} catch (InvalidTransactionException e) {
		    Trace.outln(trace, "Cannot resume transaction:"+e);
		} catch (IllegalStateException e) {
		    Trace.outln(trace, "Cannot resume transaction:"+e);
		}
	    }

	    // Call the synchronizations
	    for (int i = 0; i < synchroList.size(); i++) {
		Synchronization sync = (Synchronization) synchroList.elementAt(i);
		sync.beforeCompletion();
	    }

	    // reset context as it was before
	    if (resumed) {
		try {
		    Current.getCurrent().suspend();
		} catch (SystemException e) {
		    Trace.outln(trace, "Cannot suspend transaction:"+e);
		}
	    }
	    if (suspended) {
		try {
		    Current.getCurrent().resume(mytx);
		    resumed = true;
		} catch (SystemException e) {
		    Trace.outln(trace, "Cannot resume transaction:"+e);
		} catch (InvalidTransactionException e) {
		    Trace.outln(trace, "Cannot resume transaction:"+e);
		} catch (IllegalStateException e) {
		    Trace.outln(trace, "Cannot resume transaction:"+e);
		}
	    }
	}

	beforeCompletionDone = true;
    }

    /**
     * after completion
     */
    private void doAfterCompletion() {

	Trace.outln(trace, "SubCoordinator.doAfterCompletion()");

	// For each synchro, send afterCompletion
	for (int i = 0; i < synchroList.size(); i++) {
	    Synchronization sync = (Synchronization) synchroList.elementAt(i);
	    sync.afterCompletion(status);
	}

	// Free all connections involved in the Tx
	Current.getCurrent().freeAllConnections(tx);
    }
}
