/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.timer;

import java.util.Vector;

import org.objectweb.jonas.common.Trace;


/**
 * Clock thread for a TimerManager
 * Every second, decrement timers and launch action if expired
 */
class Clock extends Thread {

    private TimerManager tmgr;

    public Clock(TimerManager tmgr) {
	Trace.outln(TimerManager.trace, "Clock constructor");
	this.tmgr = tmgr;
    }
    
    public void run() {
	tmgr.clock();
    }
}

/**
 * Batch thread for a TimerManager
 * pocess all expired timers
 */
class Batch extends Thread {

    private TimerManager tmgr;

    public Batch(TimerManager tmgr) {
	Trace.outln(TimerManager.trace, "Batch constructor");
	this.tmgr = tmgr;
    }
    
    public void run() {
	tmgr.batch();
    }
}

/**
 * A timer manager manages 2 lists of timers with 2 threads
 * One thread is a clock which decrements timers every second
 * and passes them when expired in a list of expired timers.
 * The other thread looks in the list of expired timers to process
 * them.
 */
public class TimerManager {

    static final int trace = Trace.DB_4;

    // lists
    private Vector timerList = new Vector();
    private Vector expiredList = new Vector();

    /**
     * Constructor
     */
    public TimerManager() {
	// launch threads for timers
	new Batch(this).start();
	new Clock(this).start();
    }

    /**
     * update all timers in the list
     * each timer expired is put in a special list of expired timers
     * they will be processed then by the Batch Thread.
     */
    public void clock() {

	while (true) {
	    try {
		Thread.currentThread().sleep(1000);	// 1 second
		synchronized(timerList) {
		    int found = 0;
		    for (int i = 0; i < timerList.size(); i++) {
			TimerEvent t = (TimerEvent) timerList.elementAt(i);
			if (t.update() <= 0) {
			    timerList.removeElementAt(i--);
			    expiredList.addElement(t);
			    found++;
			}
                        // Muly changes: this is CRITICAL to the unload of the bean
                        // this is the last thing which hold the bean in memory!
                        // also the item was removed correctly from all the list there's still
                        // one more reference in this local variable which run in a seperate thread
                        // make sure we clean this variable after every iteration will do the trick!
			t = null;
		    }
		    if (found > 0) {
			timerList.notify();
		    }
		}
	    } catch (InterruptedException e) {
		Trace.errln("Timer interrupted");
	    }
	}
    }

    /**
     * process all expired timers
     */
    public void batch() {

	while (true) {
	    TimerEvent t;
	    synchronized(timerList) {
		while (expiredList.isEmpty()) {
		    try {
			timerList.wait();
		    } catch (Exception e) {
			Trace.errln("Exception in Batch: "+e);
		    }
		}
		t = (TimerEvent) expiredList.elementAt(0);
		expiredList.removeElementAt(0);
	    }
	    // Do not keep the lock during the processing of the timer
	    t.process();
	}
    }

    /**
     * add a new timer in the list
     */
    public TimerEvent addTimer(TimerEventListener tel, long timeout, Object arg) {
	TimerEvent te = new TimerEvent(tel, timeout, arg);
	synchronized(timerList) {
	    timerList.addElement(te);
	}
	return te;
    }

    /**
     * remove a timer
     * More efficient: unset directly the TimerEvent and remove it
     * later (in clock). (LATER)
     */
    public void removeTimer(TimerEvent te) {
	synchronized(timerList) {
	    timerList.removeElement(te);
	}
    }
}
