/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.tools;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.lang.reflect.Method;
import javax.ejb.deployment.ControlDescriptor;
import javax.ejb.deployment.DeploymentDescriptor;
import org.objectweb.jonas.common.Trace;
import org.objectweb.jonas.lib.DD;
import org.objectweb.jonas.lib.InvalidBeanException;
import org.objectweb.jonas.lib.JavaType;


/**
 *  This class allows to check the validity of an Enterprise Java Bean.
 */

public class CheckBean {

    static final int trace  = Trace.DB_19;

    DD ejbDD = null;
    // Classes of the interfaces and the bean
    Class  ejbRemoteClass = null;
    Class  ejbHomeClass   = null;
    Class  ejbBeanClass   = null;

    /**
     * Construct a CheckBean from a deployment descriptor of an Enterprise Java Bean.
     * @param ddesc deployment descriptor of the bean
     * @exception InvalidBeanException in fatal error case
     */
    public CheckBean(DD ddesc) throws InvalidBeanException {
	String ejbRemoteName = null;
	String ejbHomeName   = null;
	String ejbBeanName   = null;

	ejbDD = ddesc;

	// Names building
	DeploymentDescriptor deploymentDesc = null;
	deploymentDesc = ddesc.getDeploymentDescriptor();
	ejbRemoteName = deploymentDesc.getRemoteInterfaceClassName();
	if ((ejbRemoteName == null) || (ejbRemoteName.length() < 1)) {
	    throw new InvalidBeanException("The RemoteInterfaceClassName must be defined "+
					   "in the DeploymentDescriptor");
	}
	ejbHomeName = deploymentDesc.getHomeInterfaceClassName();
	if ((ejbHomeName == null) || (ejbHomeName.length() < 1)) {
	    throw new InvalidBeanException("The HomeInterfaceClassName must be defined "+
					   "in the DeploymentDescriptor");
	}
	ejbBeanName = deploymentDesc.getEnterpriseBeanClassName();
	if ((ejbBeanName == null) || (ejbBeanName.length() < 1)) {
	    throw new InvalidBeanException("The EnterpriseBeanClassName must be defined "+
					   "in the DeploymentDescriptor");
	}
	
	// Get the interfaces and the bean classes
	try {
	    ejbRemoteClass = Class.forName(ejbRemoteName);
	} catch (Exception e) {
	    Trace.errln(e, true);
	    throw new InvalidBeanException("Cannot found the Enterprise Bean's Remote Interface");
	}
	try {
	    ejbHomeClass   = Class.forName(ejbHomeName);
	} catch (Exception e) {
	    Trace.errln(e, true);
	    throw new InvalidBeanException("Cannot found the Enterprise Bean's Home interface");
	}
	try {
	    ejbBeanClass   = Class.forName(ejbBeanName);
	} catch (Exception e) {
	    Trace.errln(e, true);
	    throw new InvalidBeanException("Cannot found the Enterprise Bean's class");
	}

    }

    /**
     * Returns true if the Enterprise Java Bean is valid, false otherwise
     * @exception InvalidBeanException in fatal error case
     */
    public boolean check() throws InvalidBeanException {
	boolean isValid = true;

	if (!checkBean()) {
	    isValid = false;
	}
	if (!checkIRemote()) {
	    isValid = false;
	}
	if (!checkIHome()) {
	    isValid = false;
	}
	if (!checkDD()) {
	    isValid = false;
	}

	return(isValid);
    }

    /**
     * Returns true if the Bean class is valid, false otherwise.
     */
    private boolean checkBean() {
	boolean isValid = true;

	int modifiersClass = ejbBeanClass.getModifiers();
	if (!Modifier.isPublic(modifiersClass)) {
	    Trace.errln("Error: The Enterprise Bean's class must be defined as 'public'");
	    isValid = false;
	}
	if (Modifier.isAbstract(modifiersClass)) {
	    Trace.errln("Error: The Enterprise Bean's class must not be defined as 'abstract'");
	    isValid = false;
	}

	Field [] fieldsBean = ejbBeanClass.getFields();
	for (int i=0; i<fieldsBean.length; i++) {
	    int modField = fieldsBean[i].getModifiers();
	    if (Modifier.isStatic(modField) && !Modifier.isFinal(modField)) {
		Trace.errln("Error: All 'static' Enterprise Bean's fields must be declared as 'final'");
		isValid = false;
	    }
	}

	Constructor [] constructorsBean = ejbBeanClass.getConstructors();
	boolean isConstructorOk = false;
	for (int i=0; (i<constructorsBean.length) && (!isConstructorOk); i++) {
	    if (constructorsBean[i].getParameterTypes().length == 0) {
		if (Modifier.isPublic(constructorsBean[i].getModifiers())) {
		    isConstructorOk = true;
		}
	    }
	}
	if (!isConstructorOk) {
	    Trace.errln("Error: The Enterprise Bean's class must defined a public constructor that takes no arguments");
	    isValid = false;
	}

	return(isValid);
    }

    /**
     * Returns true if the Interface Remote class is valid, false otherwise.
     */
    private boolean checkIRemote() {
	boolean isValid = true;

	if (!ejbRemoteClass.isInterface()) {
	    Trace.errln("Error: The Enterprise Bean's Remote Interface must be an 'interface'");
	    isValid = false;
	}

	if (!JavaType.implementsInterface(ejbRemoteClass, javax.ejb.EJBObject.class)) {
	    Trace.errln("Error: The Enterprise Bean's Remote Interface must extend the " +
			"'javax.ejb.EJBObject' interface");
	    isValid = false;
	}

	Method [] methods = ejbRemoteClass.getMethods();
	for (int i=0; i<methods.length; i++) {
	    // Skip the methods defined in the 'javax.ejb.EJBObject' interface
	    Method method = methods[i];
	    if (!method.getDeclaringClass().getName().equals("javax.ejb.EJBObject")) {
		if (!checkIRemoteMethod(method)) {
		    isValid = false;
		}
	    }
	}

	return(isValid);
    }

    /**
     * Returns true if the given method of the Interface Remote class is valid,
     * false otherwise.
     */
    private boolean checkIRemoteMethod(Method method) {
	boolean isValid = true;

	if (!checkMethodThrowsRemote(method)) {
	    Trace.errln("Error: The throws clause of a method defined in the Enterprise Bean's");
	    Trace.errln("       Remote Interface must include the 'java.rmi.RemoteException'");
	    Trace.errln("  Method: '"+method.toString()+"'");
	    isValid = false;
	}

	Method [] methodsBean = ejbBeanClass.getMethods();
	boolean found = false;
	for (int i=0; (i<methodsBean.length && !found); i++) {
	    if (DD.methodsEquals(method, methodsBean[i])) {
		found = true;
		if (!checkMethodThrows(method, methodsBean[i].getExceptionTypes())) {
		    Trace.errln("Error: All the exceptions defined in the throws clause of the matching");
		    Trace.errln("       method of the Enterprise Bean class must be defined in the");
		    Trace.errln("       throws clause of the method of the Remote Interface;");
		    Trace.errln("  Method's bean: "+methodsBean[i].toString()+"'");
		    Trace.errln("  Method's remote interface: "+method.toString()+"'");
		    isValid = false;
		}
		if (!Modifier.isPublic(methodsBean[i].getModifiers())) {
		    Trace.errln("Error: A business method of an Enterprise Bean must be declared as 'public';");
		    Trace.errln("  Method's bean: "+methodsBean[i].toString()+"'");
		}
	    }
	}
	if (!found) {
	    Trace.errln("Error: A method defined in the Enterprise Bean's Remote Interface");
	    Trace.errln("       have no matching method in the Enterprise Bean's class;");
	    Trace.errln("  Method: '"+method.toString()+"'");
	    isValid = false;
	}

	return(isValid);
    }

    /**
     * Returns true if the Interface Home class is valid, false otherwise.
     */
    private boolean checkIHome() {
	boolean isValid = true;

	if (!ejbHomeClass.isInterface()) {
	    Trace.errln("Error: The Enterprise Bean's Remote Interface must be an 'interface'");
	    isValid = false;
	}

	if (!JavaType.implementsInterface(ejbHomeClass, javax.ejb.EJBHome.class)) {
	    Trace.errln("Error: The Enterprise Bean's Home Interface must extend the");
	    Trace.errln("       'javax.ejb.EJBHome' interface");
	    isValid = false;
	}

	Method [] methods = ejbHomeClass.getMethods();
	for (int i=0; i<methods.length; i++) {
	    // Skip the methods defined in the 'javax.ejb.EJBHome' interface
	    Method method = methods[i];
	    if (!method.getDeclaringClass().getName().equals("javax.ejb.EJBHome")) {
		if (!checkIHomeMethod(method)) {
		    isValid = false;
		}
	    }
	}

	return(isValid);
    }

    /**
     * Returns true if the given method of the Interface Home class is valid,
     * false otherwise.
     */
    private boolean checkIHomeMethod(Method method) {
	boolean isValid = true;

	if (!checkMethodThrowsRemote(method)) {
	    Trace.errln("Error: The throws clause of a method defined in the Enterprise Bean's");
	    Trace.errln("       Home Interface must include the 'java.rmi.RemoteException'");
	    Trace.errln("  Method: "+method.toString());
	    isValid = false;
	}

	Method [] methodsBean = ejbBeanClass.getMethods();
	if (method.getName().equals("create")) {
	    boolean found = false;
	    for (int i=0; (i<methodsBean.length && !found); i++) {
		Method mBean = methodsBean[i];
		if (mBean.getName().equals("ejbCreate")) {
		    if (DD.methodsEquals(method, mBean, "create", "create", false)) {
			found = true;
			if (!Modifier.isPublic(mBean.getModifiers())) {
			    Trace.errln("Error: The 'ejbCreate' method of the Enterprise Bean class must be");
			    Trace.errln("       declared as 'public'");
			    Trace.errln("  Method: "+mBean.toString());
			    isValid = false;
			}
			if (!checkMethodThrows(method, mBean.getExceptionTypes())) {
			    Trace.errln("Error: All the exceptions defined in the throws clause of the matching");
			    Trace.errln("       method of the Enterprise Bean class must be defined in the");
			    Trace.errln("       throws clause of the method of the Home Interface;");
			    Trace.errln("  Method's bean: "+mBean.toString()+"'");
			    Trace.errln("  Method's interface: "+method.toString()+"'");
			    isValid = false;
			}
		    }
		}
	    }
	    if (!found) {
		Trace.errln("Error: Each 'create' method of the Enterprise Bean's Home Interface");
		Trace.errln("       must match one of the 'ejbCreate' methods defined in the");
		Trace.errln("       Enterprise Bean class");
		Trace.errln("  Method: "+method.toString());
		isValid = false;
	    }

	    if (!(JavaType.getName(method.getReturnType())).equals(ejbRemoteClass.getName())) {
		Trace.errln("Error: The return type of a 'create' method of a Enterprise Bean's Home");
		Trace.errln("       Interface must be the Enterprise Bean's Remote Interface type");
		Trace.errln("  Method: "+method.toString());
		isValid = false;
	    }

	    if (!checkMethodThrowsCreate(method)) {
		Trace.errln("Error: The throws clause of a 'create' method defined in the Enterprise");
		Trace.errln("       Bean's Home Interface must include the 'javax.ejb.CreateException'");
		Trace.errln("  Method: "+method.toString());
		isValid = false;
	    }
	}

	return(isValid);
    }

    /**
     * Returns true if the deployment descriptor is valid, false otherwise
     */
    private boolean checkDD() {
	boolean isValid = true;

	ControlDescriptor[] cds = ejbDD.getDeploymentDescriptor().getControlDescriptors();
	boolean isTxBeanManaged = false;
	boolean isTxOthers      = false;
	for (int i=0; i<cds.length; i++) {
	    int txAttr = cds[i].getTransactionAttribute();
	    if (txAttr == ControlDescriptor.TX_BEAN_MANAGED) {
		isTxBeanManaged = true;
	    } else {
		isTxOthers = true;
	    }
	}
	if (isTxBeanManaged && isTxOthers) {
	    Trace.errln("Error: The TX_BEAN_MANAGED transaction attribute value must not be mixed");
	    Trace.errln("       with the other values of the transaction attributes");
	    isValid = false;
	}

	return(isValid);
    }

    /**
     * Returns true if 'java.rmi.RemoteException' is included in the throws clause
     * of the given method,
     * false otherwise.
     */
    static boolean checkMethodThrowsRemote(Method method) {
	return(checkMethodThrowsException(method, "java.rmi.RemoteException"));
    }

    /**
     * Returns true if 'javax.ejb.CreateException' is included in the throws clause
     * of the given method,
     * false otherwise.
     */
    static boolean checkMethodThrowsCreate(Method method) {
	return(checkMethodThrowsException(method, "javax.ejb.CreateException"));
    }

    /**
     * Returns true if 'javax.ejb.FinderException' is included in the throws clause
     * of the given method,
     * false otherwise.
     */
    static boolean checkMethodThrowsFinder(Method method) {
	return(checkMethodThrowsException(method, "javax.ejb.FinderException"));
    }

    /**
     * Returns true if the given exception is included in the throws clause
     * of the given method,
     * false otherwise.
     */
    static boolean checkMethodThrowsException(Method method, String exceptionName) {
	Class [] exceptions = method.getExceptionTypes();
	boolean found = false;
	for (int i=0; (i<exceptions.length && !found); i++) {
	    if (exceptions[i].getName().equals(exceptionName)) {
		found = true;
	    }
	}
	return(found);
    }

    /**
     * Returns true if the given exceptions (itself or its super-classes),
     * is included in the throws clause of the given method,
     * false otherwise.
     */
    static boolean checkMethodThrows(Method method, Class[] exceptions) {
	boolean isValid = true;

	for (int i=0; (i<exceptions.length && isValid); i++) {
	    isValid = checkMethodThrowsException(method, exceptions[i]);
	}

	return(isValid);
    }

    /**
     * Returns true if the given exception (itself or its super-classes),
     * is included in the throws clause of the given method,
     * false otherwise.
     */
    static boolean checkMethodThrowsException(Method method, Class exception) {
	Class [] exceptions = method.getExceptionTypes();
	boolean found = false;
	for (int i=0; (i<exceptions.length && !found); i++) {
	    Class superClass = exception;
	    while ((superClass != null) && !found) {
		if (exceptions[i].getName().equals(superClass.getName())) {
		    found = true;
		}
		superClass = superClass.getSuperclass();
	    }
	}
	return(found);
    }

}

