/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.tools;


import java.lang.NoSuchMethodException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import org.objectweb.jonas.common.Trace;
import org.objectweb.jonas.lib.BeanNaming;
import org.objectweb.jonas.lib.DD;
import org.objectweb.jonas.lib.InvalidBeanException;
import org.objectweb.jonas.lib.JavaType;


/**
 *  This class allows to check the validity of an Entity Enterprise Java Bean.
 */

public class CheckEntity extends CheckBean {

    static final int trace  = Trace.DB_19;

    Class ejbPrimaryKeyClass = null;

    /**
     * Construct a CheckEntity from a deployment descriptor of an Entity Enterprise Java Bean.
     * @param ddesc deployment descriptor of the entity bean
     * @exception InvalidBeanException in fatal error case
     */
    public CheckEntity(DD ddesc) throws InvalidBeanException {

	super(ddesc);
	
	String ejbPrimaryKeyName = null;

	ejbPrimaryKeyName = (ddesc.getEntityDescriptor()).getPrimaryKeyClassName();
	if ((ejbPrimaryKeyName == null) || (ejbPrimaryKeyName.length() < 1)) {
	    throw new InvalidBeanException("The 'PrimaryKeyClassName' must be defined "+
					   "in the EntityDescriptor");
	}
	try {
	    ejbPrimaryKeyClass = Class.forName(ejbPrimaryKeyName);
	} catch (Exception e) {
	    Trace.errln(e, true);
	    throw new InvalidBeanException("Cannot found the Enterprise Bean's Primary Key class");
	}

    }

    /**
     * Returns true if the Entity Bean is valid, false otherwise
     * @exception InvalidBeanException in fatal error case
     */
    public boolean check() throws InvalidBeanException {
	boolean isValid = true;
	if (!super.check()) {
	    isValid = false;
	}
	if (!checkIHome()) {
	    isValid = false;
	}
	if (!checkPrimaryKey()) {
	    isValid = false;
	}
	if (!checkBean()) {
	    isValid = false;
	}
	return(isValid);
    }

    /**
     * Returns true if the Interface Home class is valid, false otherwise.
     */
    private boolean checkIHome() {
	boolean isValid = true;

	boolean isFindByPK = false;
	Method [] methods = ejbHomeClass.getMethods();
	for (int i=0; i<methods.length; i++) {
	    // Skip the methods defined in the 'javax.ejb.EJBHome' interface
	    Method method = methods[i];
	    if (!method.getDeclaringClass().getName().equals("javax.ejb.EJBHome")) {
		if (method.getName().equals("findByPrimaryKey")) {
		    isFindByPK = true;
		}
		if (!checkIHomeMethod(method)) {
		    isValid = false;
		}
	    }
	}
	if (!isFindByPK) {
	    Trace.errln("Error: The Home Interface of every Entity Bean includes the");
	    Trace.errln("       'findByPrimaryKey' method");
	    isValid = false;
	}

	return(isValid);
    }

    /**
     * Returns true if the given method of the Interface Home class is valid,
     * false otherwise.
     */
    private boolean checkIHomeMethod(Method method) {
	boolean isValid = true;

	boolean isCreate    = method.getName().equals("create");
	boolean isFindByPK  = method.getName().equals("findByPrimaryKey");
	boolean isFindOther = ((!isFindByPK) && (method.getName().indexOf("find") == 0));
	String  methodBeanName = null;

	if (isCreate) {
	    methodBeanName = new String("ejbCreate");
	} else if (isFindByPK) {
	    methodBeanName = new String("ejbFindByPrimaryKey");
	} else if (isFindOther) {
	    methodBeanName = "ejb" + BeanNaming.firstToUpperCase(method.getName());
	} else {
	    Trace.errln("Error: Only 'create' and 'finder' methods can be defined in an");
	    Trace.errln("       Entity Bean's Home Interface");
	    Trace.errln("  Method: "+method.toString());
	    isValid = false;
	}

	// Found the associated method in the Bean class
	Method [] methodsBean = ejbBeanClass.getMethods();
	boolean   found = false;
	Method    methodBean = null;
	for (int i=0; (i<methodsBean.length && !found); i++) {
	    methodBean = methodsBean[i];
	    if (methodBean.getName().equals(methodBeanName)) {
		if (DD.methodsEquals(method, methodBean, method.getName(), 
				     method.getName(), false)) {
			found = true;
		}
	    }
	}

	if (isCreate) {
	    // Partially checked in CheckBean - Here is the specific checking for Entity
	    if (found) {
		if (ejbDD.isEntityContainerManagedPersistence()) {
		    if (!JavaType.isVoid(methodBean.getReturnType())) {
			Trace.errln("Error: The return value of an 'ejbCreate' method must be 'void'");
			Trace.errln("       for Enterprise Beans with container-managed persistence");
			Trace.errln("  Method: "+methodBean.toString());
			isValid = false;
		    }
		} else {
		    if (!(JavaType.getName(methodBean.getReturnType())).equals(JavaType.getName(ejbPrimaryKeyClass))) {
			Trace.errln("Error: The return type of an 'ejbCreate' method must be the");
			Trace.errln("       Primary Key type for Enterprise Beans with bean-managed");
			Trace.errln("       persistence");
			Trace.errln("  Method: "+methodBean.toString());
			isValid = false;	
		    }
		}
	    }
	    // Check for ejbPostCreate()
	    methodBeanName = new String("ejbPostCreate");
	    found = false;
	    methodBean = null;
	    for (int i=0; (i<methodsBean.length && !found); i++) {
		methodBean = methodsBean[i];
		if (methodBean.getName().equals(methodBeanName)) {
		    if (DD.methodsEquals(method, methodBean, method.getName(), 
					 method.getName(), false)) {
			found = true;
			if (!Modifier.isPublic(methodBean.getModifiers())) {
			    Trace.errln("Error: The 'ejbPostCreate' method of the Enterprise Bean class must be");
			    Trace.errln("       declared as 'public'");
			    Trace.errln("  Method: "+methodBean.toString());
			    isValid = false;
			}
			if (!JavaType.isVoid(methodBean.getReturnType())) {
			    Trace.errln("Error: The return value of an 'ejbPostCreate' method must be 'void'");
			    Trace.errln("       for Enterprise Beans with container-managed persistence");
			    Trace.errln("  Method: "+methodBean.toString());
			    isValid = false;
			}
			if (!checkMethodThrows(method, methodBean.getExceptionTypes())) {
			    Trace.errln("Error: All the exceptions defined in the throws clause of the matching");
			    Trace.errln("       method of the Enterprise Bean class must be defined in the");
			    Trace.errln("       throws clause of the method of the Home Interface");
			    Trace.errln("  Method's bean: "+methodBean.toString());
			    Trace.errln("  Method's interface: "+method.toString());
			    isValid = false;
			}
		    }
		}
	    }
	    if (!found) {
		Trace.errln("Error: Each 'create' method of the Enterprise Bean's Home Interface");
		Trace.errln("       must match one of the 'ejbPostCreate' methods defined in the");
		Trace.errln("       Enterprise Bean class");
		Trace.errln("  Method: "+method.toString());
		isValid = false;
	    }
	} 

	if (isFindByPK || isFindOther) {
	    if (!checkMethodThrowsFinder(method)) {
		Trace.errln("Error: The throws clause of a 'finder<Method>' method defined in the");
		Trace.errln("       Enterprise Bean's Home Interface must include the");
		Trace.errln("       'javax.ejb.FinderException'");
		Trace.errln("  Method: "+method.toString());
		isValid = false;
	    }
	    if (found) {
		if (!checkMethodThrows(method, methodBean.getExceptionTypes())) {
		    Trace.errln("Error: All the exceptions defined in the throws clause of the matching");
		    Trace.errln("       method of the Enterprise Bean class must be defined in the");
		    Trace.errln("       throws clause of the method of the Home Interface");
		    Trace.errln("  Method's bean: "+methodBean.toString());
		    Trace.errln("  Method's interface: "+method.toString());
		    isValid = false;
		}
		if (!Modifier.isPublic(methodBean.getModifiers())) {
		    Trace.errln("Error: The 'ejbFind<Method>' method of the Enterprise Bean class must be");
		    Trace.errln("       declared as 'public'");
		    Trace.errln("  Method: "+methodBean.toString());
		    isValid = false;
		}
	    }
	    if (ejbDD.isEntityContainerManagedPersistence() && found) {
		Trace.errln("Error: The enterprise Bean provider does not write the finder");
		Trace.errln("       'ejbFind<Method>' methods for Entity Bean with");
		Trace.errln("       container-managed persistence");
		Trace.errln("  Method: "+methodBean.toString());
		isValid = false;  
	    }
	    if (!ejbDD.isEntityContainerManagedPersistence() && !found) {
		Trace.errln("Error: The Enterprise Bean class must define the 'ejbFinder<Method>'");
		Trace.errln("       method associated to");
		Trace.errln("       "+method.toString());
		isValid = false;
	    }
	}

	if (isFindByPK) {
	    if (!(JavaType.getName(method.getReturnType())).equals(ejbRemoteClass.getName())) {
		Trace.errln("Error: The return type of a 'findByPrimary' method must be the");
		Trace.errln("       Enterprise Bean's Remote Interface type");
		Trace.errln("  Method: "+method.toString());
		isValid = false;
	    }
	    Class [] params = method.getParameterTypes();
	    boolean paramOk = true;
	    if (params.length != 1) {
		paramOk = false;
	    } else if (!(JavaType.getName(params[0])).equals(ejbPrimaryKeyClass.getName())) {
		paramOk = false;
	    }
	    if (!paramOk) {
		Trace.errln("Error: The 'findByPrimary' method must have only one parameter");
		Trace.errln("       with Primary Key type");
		Trace.errln("  Method: "+method.toString());
		isValid = false;
	    }
	}
	
	if (isFindOther) {
	    String returnTypeName = JavaType.getName(method.getReturnType());
	    if (!returnTypeName.equals(ejbRemoteClass.getName()) &&
		!returnTypeName.equals("java.util.Enumeration") &&
		!returnTypeName.equals("java.util.Collection")) {
		Trace.errln("Error: The return type of a 'finder' method must be the Enterprise");
		Trace.errln("       Bean's Remote Interface type or a collection of thereof");
		Trace.errln("  Method: "+method.toString());
		isValid = false;
	    }
	}

	return(isValid);
    }

    /**
     * Returns true if the Primary Key class is valid, false otherwise.
     */
    private boolean checkPrimaryKey() {
	boolean isValid = true;

	if (!JavaType.isSerializable(ejbPrimaryKeyClass)) {
	    Trace.errln("Error: The Primary Key class must implement a toString() method");
	    Trace.errln("       programming language Serialization protocol");
	    Trace.errln("  Class: "+ejbPrimaryKeyClass.getName());
	    isValid = false;
	}

	try {
	    Method method = ejbPrimaryKeyClass.getDeclaredMethod("toString", new Class[0]);
	} catch (NoSuchMethodException e) {
	    Trace.errln("Error: A toString() method must be defined in the Primary Key class");
	    isValid = false;
	}
	
	return(isValid);
    }

    /**
     * Returns true if the Entity Bean class is valid, false otherwise.
     */
    private boolean checkBean() {
	boolean isValid = true;

	if (!JavaType.implementsInterface(ejbBeanClass, javax.ejb.EntityBean.class)) {
	    Trace.errln("Error: The Enterprise Entity Bean class must implement the");
	    Trace.errln("       'javax.ejb.EntityBean' interface.");
	    isValid = false;
	}

	return(isValid);
    }

}

