/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.tools;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import javax.ejb.deployment.EntityDescriptor;
import org.objectweb.jonas.common.Trace;
import org.objectweb.jonas.lib.DD;
import org.objectweb.jonas.lib.InvalidBeanException;
import org.objectweb.jonas.lib.JavaType;


/**
 *  This class allows to check the validity of an Entity Enterprise Java Bean
 *  with Container Managed Persistence.
 *  Its extends CheckEntity class, itself extends CheckBean class.
 */

public class CheckEntityCM extends CheckEntity {

    static final int trace  = Trace.DB_19;

    /**
     * Construct a CheckEntityCM from a deployment descriptor of an
     * Entity Enterprise Java Bean with container managed persistence.
     * @param ddesc deployment descriptor of the entity bean
     * @exception InvalidBeanException in fatal error case
     */
    public CheckEntityCM(DD ddesc) throws InvalidBeanException {
	super(ddesc);
	if (!ejbDD.isEntityContainerManagedPersistence()) {
	    throw new InvalidBeanException("An Entity Bean with container-managed persistence "+
					   "must have the container-managed fields "+
					   "defined in the Entity Descriptor.");
	}
    }

    /**
     * Returns true if the Entity Bean with container managed persistence is valid,
     * false otherwise.
     * @exception InvalidBeanException in fatal error case
     */
    public boolean check() throws InvalidBeanException {
	boolean isValid = true;
	if (!super.check()) {
	    isValid = false;
	}
	if (!checkPrimaryKey()) {
	    isValid = false;
	}
	if (!checkBean()) {
	    isValid = false;
	}
	return(isValid);
    }

    /**
     * Returns true if the Primary Key class is valid, false otherwise.
     */
    private boolean checkPrimaryKey() {
	boolean isValid = true;

	if (!Modifier.isPublic(ejbPrimaryKeyClass.getModifiers())) {
	    Trace.errln("Error: The Primary Key type must be 'public' for an Entity Bean with");
	    Trace.errln("       container-managed persistence");
	    Trace.errln("  Class: "+ejbPrimaryKeyClass.toString());
	    isValid = false;
	}

	Field [] fieldsPK = ejbPrimaryKeyClass.getDeclaredFields();
	Field [] fieldsCM = ejbDD.getEntityDescriptor().getContainerManagedFields();
	for (int i=0; i<fieldsPK.length; i++) {
	    if (!Modifier.isPublic(fieldsPK[i].getModifiers())) {
		Trace.errln("Error: All fields of the Primary Key class must be declared as 'public'");
		Trace.errln("       for an Entity Bean with container-managed persistence");
		Trace.errln("  Field: "+fieldsPK[i].toString());
		isValid = false;
	    }
	    boolean isFieldContainerManaged = false;
	    String fieldPKName = fieldsPK[i].getName();
	    for (int j=0; (j<fieldsCM.length && !isFieldContainerManaged); j++) {
		if (fieldPKName.equals(fieldsCM[j].getName())) {
		    isFieldContainerManaged = true;
		}
	    }
	    if (!isFieldContainerManaged) {
		Trace.errln("Error: The names of the fields in the Primary Key class must be a subset");
		Trace.errln("       of the names of the container-managed fields of the Entity Bean");
		Trace.errln("  Field of the Primary Key: "+fieldsPK[i].toString());
		isValid = false;
	    }
	}

	Constructor [] constructors = ejbPrimaryKeyClass.getConstructors();
	boolean isDefaultConstructor = false;
	for (int i=0; (i<constructors.length && !isDefaultConstructor); i++) {
	    Class [] params = constructors[i].getParameterTypes();
	    if ((params.length == 0) && Modifier.isPublic(constructors[i].getModifiers())) {
		isDefaultConstructor = true;
	    }
	}
	if (!isDefaultConstructor) {
	    Trace.errln("Error: The Primary Key class must have a 'public' default constructor");
	    Trace.errln("       for an Entity Bean with container-managed persistence");
	    Trace.errln("  Class: "+ejbPrimaryKeyClass.toString());
	    isValid = false;
	}

	return(isValid);
    }

    /**
     * Returns true if the Entity Bean class is valid, false otherwise.
     */
    private boolean checkBean() {
	boolean isValid = true;

	Field [] fieldsCM = ejbDD.getEntityDescriptor().getContainerManagedFields();
	for (int i=0; i<fieldsCM.length; i++) {
	    if (!Modifier.isPublic(fieldsCM[i].getModifiers())) {
		Trace.errln("Error: The container-managed field must be declared as 'public'");
		Trace.errln("  Field: "+fieldsCM[i].toString());
		isValid = false;
	    }
	    if (!JavaType.isSerializable(fieldsCM[i].getType())) {
		Trace.errln("The container-managed field must be of a Java Serializable type");
		Trace.errln("  Field: "+fieldsCM[i].toString());
		isValid = false;
	    }
	}

	return(isValid);
    }

}

