/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.tools;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.Properties;
import org.objectweb.jonas.common.Trace;
import org.objectweb.jonas.lib.DD;
import org.objectweb.jonas.lib.InvalidBeanException;
import org.objectweb.jonas.lib.JavaType;


/**
 *  This class allows to check the validity of an Entity Enterprise Java Bean
 *  with Container Managed Persistence on JDBC.
 *  Its extends CheckEntityCM class, itself extends CheckEntity class,
 *  itself extends CheckBean class.
 */

public class CheckEntityCMbyJDBC extends CheckEntityCM {

    static final int trace  = Trace.DB_19;

    /**
     * Construct a CheckEntityCMbyJDBC from a deployment descriptor of an
     * Entity Enterprise Java Bean with container managed persistence by JDBC.
     * @param ddesc deployment descriptor of the entity bean
     * @exception InvalidBeanException in fatal error case
     */
    public CheckEntityCMbyJDBC(DD ddesc) throws InvalidBeanException {
	super(ddesc);
    }

    /**
     * Returns true if the Entity Bean with container managed persistence by JDBC is valid,
     * false otherwise.
     * @exception InvalidBeanException in fatal error case
     */
    public boolean check() throws InvalidBeanException {
	boolean isValid = true;
	if (!super.check()) {
	    isValid = false;
	}
	if (!checkDDProperties()) {
	    isValid = false;
	}
	if (!checkBean()) {
	    isValid = false;
	}
	return(isValid);
    }

    /**
     * Returns true if the Properties defined in the Deployment Descriptor are valid,
     * false otherwise.
     */
    private boolean checkDDProperties() {
	boolean isValid = true;

	Properties props  = null;
	props = ejbDD.getDeploymentDescriptor().getEnvironmentProperties();

	if (props.getProperty("datasource.name") == null) {
	    Trace.errln("Error: For an Entity Bean with container-managed persistence,");
	    Trace.errln("       the property 'datasource.name' must be defined in the Enterprise Bean's");
	    Trace.errln("       environment properties of the Deployment Descriptor");
	    isValid = false;
	}

	if (props.getProperty("db.TableName") == null) {
	    Trace.errln("Error: For an Entity Bean with container-managed persistence,");
	    Trace.errln("       the property 'db.TableName' must be defined in the Enterprise Bean's");
	    Trace.errln("       environment properties of the Deployment Descriptor");
	    isValid = false;
	}

	Field [] fieldsCM = ejbDD.getEntityDescriptor().getContainerManagedFields();
	for (int i=0; i<fieldsCM.length; i++) {
	    String propName = "db.Field."+fieldsCM[i].getName();
	    if (props.getProperty(propName) == null) {
		Trace.errln("Error: For an Entity Bean with container-managed persistence,");
		Trace.errln("       the property '"+propName+"' must be defined");
		Trace.errln("       in the Enterprise Bean's environment properties of the");
		Trace.errln("       Deployment Descriptor");
		isValid = false;
	    }
	}

	Method [] methsHome = ejbHomeClass.getMethods();
	for (int i=0; i<methsHome.length; i++) {
	    // Skip the methods defined in the 'javax.ejb.EJBHome' interface
	    Method methHome = methsHome[i];
	    if (!methHome.getDeclaringClass().getName().equals("javax.ejb.EJBHome")) {
		String methName = methHome.getName();
		if (!methName.equals("findByPrimaryKey") && (methName.indexOf("find") == 0)) {
		    String propName = "db.Finder."+methName;
		    if (props.getProperty(propName) == null) {
			Trace.errln("Error: For an Entity Bean with container-managed persistence,");
			Trace.errln("       the property '"+propName+"' must be defined");
			Trace.errln("       in the Enterprise Bean's environment properties of the");
			Trace.errln("       Deployment Descriptor");
			isValid = false;
		    }
		}
	    }
	}

	return(isValid);
    }

    /**
     * Returns true if the Entity Bean class is valid, false otherwise.
     */
    private boolean checkBean() {
	boolean isValid = true;

	Field [] fieldsCM = ejbDD.getEntityDescriptor().getContainerManagedFields();
	for (int i=0; i<fieldsCM.length; i++) {
	    Class typeField = fieldsCM[i].getType();
	    if ((JavaType.getSQLGetMethod(typeField) == null) &&
		(JavaType.getSQLSetMethod(typeField) == null)) {
		Trace.errln("Error: Cannot container-managed a field because of its type");
		Trace.errln("  Field: "+fieldsCM[i].toString());
		isValid = false;
	    }
	}

	return(isValid);
    }

}

