/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.tools;


import java.io.IOException;
import java.lang.String;
import java.lang.reflect.Method;
import javax.ejb.deployment.DeploymentDescriptor;
import org.objectweb.jonas.common.IndentedFile;
import org.objectweb.jonas.common.Trace;
import org.objectweb.jonas.lib.BeanNaming;
import org.objectweb.jonas.lib.DD;
import org.objectweb.jonas.lib.JavaType;


/**
 *  This class allows to generate the source of the class that implements
 *  the Enterprise bean's Home interface of a given Enterprise Java Bean.
 */
abstract class GenICHome {

    static final int trace = Trace.DB_19;

    IndentedFile src = null;

    DD ddesc = null;

    // Full names of the interfaces and the bean classe
    String ejbRemoteName = null;
    String ejbHomeName   = null;
    String ejbBeanName   = null;
    // Classes of the interfaces and the bean
    Class  ejbRemoteClass = null;
    Class  ejbHomeClass   = null;
    Class  ejbBeanClass   = null;
    // Names of the implementation classes (Full names and base names)
    String wrpRemoteName     = null;
    String wrpHomeName       = null;
    String wrpRemoteBaseName = null;
    String wrpHomeBaseName   = null;

    /**
     * GenICHome Constructor
     *
     * @param dd       deployment descriptor of the Enterprise Java Bean
     * @param fileName name of the source file of the class to generate
     * @exception GenICException in error case
     */
    GenICHome(DD dd, String fileName) throws GenICException {
	// Source File
	try {
	    src = new IndentedFile(fileName);
	} catch (IOException e) {
	    Trace.errln(e, true);
	    throw new GenICException("Cannot create '" + fileName + "' file");
	}
	// Deployment Descriptor
	ddesc = dd;
	// Names building
	DeploymentDescriptor deploymentDesc = ddesc.getDeploymentDescriptor();
	ejbRemoteName = deploymentDesc.getRemoteInterfaceClassName();
	if ((ejbRemoteName == null) || (ejbRemoteName.length() < 1)) {
	    throw new GenICException("The RemoteInterfaceClassName MUST BE defined "+
				     "in the DeploymentDescriptor");
	}
	ejbHomeName = deploymentDesc.getHomeInterfaceClassName();
	if ((ejbHomeName == null) || (ejbHomeName.length() < 1)) {
	    throw new GenICException("The HomeInterfaceClassName MUST BE defined "+
				     "in the DeploymentDescriptor");
	}
	ejbBeanName = deploymentDesc.getEnterpriseBeanClassName();
	if ((ejbBeanName == null) || (ejbBeanName.length() < 1)) {
	    throw new GenICException("The EnterpriseBeanClassName MUST BE defined "+
				     "in the DeploymentDescriptor");
	}
	String sPack;
	BeanNaming bn     = new BeanNaming(deploymentDesc);
	wrpRemoteBaseName = bn.getWrpRemoteName();
	wrpRemoteName     = bn.getFullWrpRemoteName();
	wrpHomeBaseName   = bn.getWrpHomeName();
	wrpHomeName       = bn.getFullWrpHomeName();
	// Get the interfaces and the bean classes
	try {
	    ejbRemoteClass = Class.forName(ejbRemoteName);
	    ejbHomeClass   = Class.forName(ejbHomeName);
	    ejbBeanClass   = Class.forName(ejbBeanName);
	} catch (Exception e) {
	    Trace.errln(e, true);
	    throw new GenICException("Cannot found enterprise Bean's class");
	}
    }

    /**
     * Generate the source of the class which implements the Home interface
     * @param withError the source file is generated with a syntax error
     *                  (internal use only for test)
     * @exception GenICException in error case
     */
    void generate(boolean withError) throws GenICException {

	// Declared methods in the Enterprise bean's home interface;
	Method methods[];

	// Comments
	src.println("//");
	src.println("// "+wrpHomeBaseName+".java");
	src.println("// DO NOT EDIT THIS FILE");
	src.println("// It is automatically generated via the GenIC tool");
	src.println("//");
	src.println();

	// Package clause
	String s = BeanNaming.getPackageName(ejbHomeName);
	if (s.length() != 0) {
	    src.println("package "+s+";");
	}
	src.println();

	// Wrong import clause generation for test only !!!!!
	if (withError) {
	    src.println("import java.toto.Titi;");
	    src.println();
	}

	// Import clauses
	genImport();
	src.println();

	// Class declaration
	src.println("public class "+wrpHomeBaseName+
		    " extends "+getSuperClass()+
		    " implements "+ejbHomeName+" {");
	src.println();
	src.indentPlus();

	// Constructor()
	genConstructor();
	src.println();

	// create() and find...() Methodes defined in the home interface
	methods = ejbHomeClass.getMethods();
	for (int m=0; m < methods.length; m++) {
	    // Skip the methods defined in the 'javax.ejb.EJBHome' interface
	    Method method = methods[m];
	    if (!method.getDeclaringClass().getName().equals("javax.ejb.EJBHome")) {

		boolean isCreate    = method.getName().equals("create");
		boolean isFindByPK  = method.getName().equals("findByPrimaryKey");
		boolean isFindOther = ((!isFindByPK) && (method.getName().indexOf("find") == 0));

		if (isCreate) {
		    genCreate(method);
		} else if (isFindByPK) {
		    genFindByPK(method);
		} else if (isFindOther) {
		    genFindOther(method);
		} else {
		    Trace.outln("GenICHome() WARNING: Method '" + method.getName() +
				"' in the Enterprise bean's home interface IGNORED");
		}

		src.println();
	    }
	}

	// Class End
	src.indentMinus();
	src.println("}");
	src.println();

	//
	src.close();
    }

    /**
     * Source generation of the header of the given Method.
     * (Declaration and the begining of the core..)
     * @exception GenICException in error case
     */
    void genHeaderMethod(Method method) throws GenICException {

	Class paramsTable[]     = method.getParameterTypes();
	Class exceptionsTable[] = method.getExceptionTypes();

	// declaration
	src.print("synchronized public " +
		  JavaType.getName(method.getReturnType()) +
		  " " +
		  method.getName() +
		  "(");
	for (int p=1; p <= paramsTable.length; p++) {
	    src.print(JavaType.getName(paramsTable[p-1]) + " " + "p" + p, false);
	    if (p < paramsTable.length) {
		src.print(", ", false);
	    }
	}
	src.print(") ", false);
	for (int e=1; e <= exceptionsTable.length; e++) {
	    if (e==1) {
		src.print("throws ", false);
	    }
	    src.print(exceptionsTable[e-1].getName(), false);
	    if (e < exceptionsTable.length) {
		src.print(",", false);
	    }
	    src.print(" ", false);
	}
	src.println("{", false);
	// begining of the core
	src.indentPlus();
	src.print("Trace.outln(Trace.DB_15, \""+wrpHomeBaseName+"."+method.getName()+"(");
	for (int p=1; p <= paramsTable.length; p++) {
	    src.print(JavaType.getName(paramsTable[p-1]), false);
	    if (p < paramsTable.length) {
		src.print(", ", false);
	    }
	}
	src.println(")\");", false);
    }

    /**
     * Source generation of the 'catch' code following any call of a bean method
     * (such bussiness methods, ejbCreate(), ejbPostCreate(),, .....)
     * The code generated may be:
     *     } catch (RemoteException e) {
     *         ....
     *     } catch (RuntimeException e) {
     *         ....
     *     } catch (Error e) {
     *         ....
     * See the chapter 12.3 in the EJB Specification document
     * @param mInter    the current generated method of the Home of the Remote interface
     * @param mBeanName the name of the associated method of the bean
     *                  (i.e. ejbCreate() or ejbPostCreate() for a create() interface method, ...)
     */
    void genCatchBeanException(Method mInter, String mBeanName)
	throws GenICException {
	// The "try { ....;" has already been generated
	// The "}" will be generate

	// See if the RemoteException must be catched
	// The RemoteException must be catched if it is in the throws clause
	// of the method of the bean.
	boolean catchRemote = false;
	boolean found = false;
	Method[] methodsBean = ejbBeanClass.getMethods();
	Method mBean = null;
	for (int i=0; ((i<methodsBean.length) && !found); i++) {
	    found = DD.methodsEquals(mInter, methodsBean[i], mBeanName, methodsBean[i].getName());
	    if (found) {
		mBean = methodsBean[i];
	    }
	}
	if (found) {
	    Class[] exceptions = mBean.getExceptionTypes();
	    for (int i=0; ((i<exceptions.length) && (!catchRemote)); i++) {
		catchRemote = ((exceptions[i]).getName()).equals("java.rmi.RemoteException");
	    }
	} else {
	    Trace.outln("GenICHome() WARNING: Method '" + mBeanName +
			"' associated to the interface method '" +
			mInter + "' not found in the Enterprise Bean");
	}

	// Source Generation ...
	src.indentMinus();
	// Catch the System-level exceptions java.rmi.RemoteException
	// See EJB-Specification page 109, &12.3.2
	if(catchRemote){
	    src.println("} catch (RemoteException e) {");
	    src.indentPlus();
	    src.println("Trace.errln(\"Get the following RemoteException thrown by an enterprise Bean's method (rethrow)\");");
	    src.println("Trace.errln(e);");
	    src.println("throw e;");
	    src.indentMinus();
	}
	// Catch the Unchecked exceptions java.lang.RuntimeException and java.lang.Error
	// See EJB-Specification page 109, &12.3.3
	src.println("} catch (RuntimeException e) {");
	src.indentPlus();
	src.println("Trace.errln(\"Get the following RuntimeException thrown by an enterprise Bean's method (convert to a java.rmi.RemoteException)\");");
	src.println("Trace.errln(e);");
	src.println("onUncheckedException(e);");
	src.println("throw new RemoteException(\"Get '\" + e + \"' thrown by an enterprise Bean's method\", e);");
	src.indentMinus();
	src.println("} catch (Error e) {");
	src.indentPlus();
	src.println("Trace.errln(\"Get the following Error thrown by an enterprise Bean's method (convert to a java.rmi.RemoteException)\");");
	src.println("Trace.errln(e);");
	src.println("onUncheckedException(e);");
	src.println("throw new RemoteException(\"Get '\" + e + \"' thrown by an enterprise Bean's method\", e);");

    }

    /**
     * Name of the superclass of the generated class
     * (JSessionHome or JEntityHome, depending on the Bean type)
     */
    abstract String getSuperClass();

    /**
     * Source generation of the import clauses
     * @exception GenICException in error case
     */
    abstract void   genImport() throws GenICException;

    /**
     * Source generation of the constructor method
     * @exception GenICException in error case
     */
    abstract void   genConstructor() throws GenICException;

    /**
     * Source generation for the given create() method described in the Home interface
     * @exception GenICException in error case
     */
    abstract void   genCreate(Method method) throws GenICException;

    /**
     * Source generation for the given findByPrimaryKey() method described in the Home interface
     * @exception GenICException in error case
     */
    abstract void   genFindByPK(Method method) throws GenICException;

    /**
     * Source generation for the given findXXX() method described in the Home interface
     * @exception GenICException in error case
     */
    abstract void   genFindOther(Method method) throws GenICException;

}

