/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.tools;


import java.io.IOException;
import java.lang.String;
import java.lang.reflect.Method;
import javax.ejb.deployment.DeploymentDescriptor;
import org.objectweb.jonas.common.IndentedFile;
import org.objectweb.jonas.common.Trace;
import org.objectweb.jonas.lib.BeanNaming;
import org.objectweb.jonas.lib.DD;
import org.objectweb.jonas.lib.JavaType;


/**
 *  This class allows to generate the source of the class that implements
 *  the Enterprise bean's remote interface of a given Enterprise Java Bean.
 */

abstract class GenICRemote {

    IndentedFile src   = null;

    DD           ddesc = null;

    // Full names of the interfaces and the bean classe
    String ejbRemoteName     = null;
    String ejbBeanName       = null;
    // Classes of the interfaces and the bean
    Class  ejbRemoteClass = null;
    Class  ejbBeanClass   = null;
    // Names of the implementation classes (Full names and base names)
    String wrpRemoteName     = null;
    String wrpRemoteBaseName = null;


    /**
     * GenICRemote Constructor
     *
     * @param dd       deployment descriptor of the Enterprise Java Bean
     * @param fileName name of the source file of the class to generate
     * @exception GenICException in error case
     */
    GenICRemote(DD dd, String fileName) throws GenICException {
	// Source File
	try {
	    src = new IndentedFile(fileName);
	} catch (IOException e) {
	    Trace.errln(e, true);
	    throw new GenICException("Cannot create '" + fileName + "' file");
	}
	// Deployment Descriptor
	ddesc = dd;
	// Names building
	DeploymentDescriptor deploymentDesc = ddesc.getDeploymentDescriptor();
	ejbRemoteName = deploymentDesc.getRemoteInterfaceClassName();
	if ((ejbRemoteName == null) || (ejbRemoteName.length() < 1)) {
	    throw new GenICException("The RemoteInterfaceClassName MUST BE defined "+
				     "in the DeploymentDescriptor");
	}
	ejbBeanName = deploymentDesc.getEnterpriseBeanClassName();
	if ((ejbBeanName == null) || (ejbBeanName.length() < 1)) {
	    throw new GenICException("The EnterpriseBeanClassName MUST BE defined "+
				     "in the DeploymentDescriptor");
	}
	String sPack;
	BeanNaming bn     = new BeanNaming(deploymentDesc);
	wrpRemoteBaseName = bn.getWrpRemoteName();
	wrpRemoteName     = bn.getFullWrpRemoteName();
	// Get the interfaces and the bean classes
	try {
	    ejbRemoteClass = Class.forName(ejbRemoteName);
	    ejbBeanClass   = Class.forName(ejbBeanName);
	} catch (Exception e) {
	    Trace.errln(e, true);
	    throw new GenICException("Cannot found enterprise Bean's class");
	}

    }

    /**
     * Generate the source of the class which implements the Remote interface
     * @param withError the source file is generated with a syntax error
     *                  (internal use only for test)
     * @exception GenICException in error case
     */
    void generate(boolean withError) throws GenICException {

	// Declared methods in the Enterprise bean's remote interface;
	Method methods[];

	// Comments
	src.println("//");
	src.println("// "+wrpRemoteBaseName+".java");
	src.println("// DO NOT EDIT THIS FILE");
	src.println("// It is automatically generated via the GenIC tool");
	src.println("//");
	src.println();

	// Package clause
	String s = BeanNaming.getPackageName(ejbRemoteName);
	if (s.length() != 0) {
	    src.println("package "+s+";");
	}
	src.println();

	// Wrong import clause generation for test only !!!!!
	if (withError) {
	    src.println("import java.toto.Titi;");
	    src.println();
	}

	// Import clauses
	genImport();
	src.println();

	// Class declaration
	src.println("public class "+wrpRemoteBaseName+
		    " extends "+getSuperClass()+
		    " implements "+ejbRemoteName+" {");
	src.println();
	src.indentPlus();

	// Constructor()
	genConstructor();
	src.println();

	// getHandle() Method
	genGetHandle();
	src.println();

	// Bussiness Methodes defined in the remote interface
	methods = ejbRemoteClass.getMethods();
	for (int m=0; m < methods.length; m++) {
	    // Skip the methods defined in the 'javax.ejb.EJBObject' interface
	    Method method = methods[m];
	    if (!method.getDeclaringClass().getName().equals("javax.ejb.EJBObject")) {
		genBussinessMethod(method);
		src.println();
	    }
	}

	// createData(), removeData(), loadData(), storeData()
	genAccessData();
	src.println();

	// End of the Class
	src.indentMinus();
	src.println("}");
	src.println();

	//
	src.close();
    }

    /**
     * Source generation for the given bussiness method described in the Remote interface
     * @exception GenICException in error case
     */
    void genBussinessMethod(Method method) throws GenICException {

	int transAttribute;
	Class params[] = method.getParameterTypes();
	Class exceptions[] = method.getExceptionTypes();
	boolean isReturnValue = ! JavaType.isVoid(method.getReturnType());

	// declaration
	src.print("public ");
	if (ddesc.isSession()) {
	    src.print("synchronized ", false);
	}
	src.print(JavaType.getName(method.getReturnType()) + " " +
		  method.getName() +
		  "("
		  , false);

	for (int p=1; p <= params.length; p++) {
	    src.print(JavaType.getName(params[p-1]) + " " + "p" + p, false);
	    if (p < params.length) {
		src.print(", ", false);
	    }
	}
	src.print(") ", false);
	for (int e=1; e <= exceptions.length; e++) {
	    if (e==1) {
		src.print("throws ", false);
	    }
	    src.print(exceptions[e-1].getName(), false);
	    if (e < exceptions.length) {
		src.print(",", false);
	    }
	    src.print(" ", false);
	}
	src.println("{", false);

	// core
	src.indentPlus();
	src.println("Trace.outln(Trace.DB_15, \"" +
		    wrpRemoteBaseName + "." + method.getName() +
		    "()\");");
	src.println(ejbBeanName + " eb = null;");
	try {
	    transAttribute = ddesc.getTransactionAttributeForMethod(method);
	} catch (Exception e) {
	    Trace.errln(e, true);
	    throw new GenICException("There is no transactionAttribute for the method '" +
				     method.toString() + "'");
	}
	src.println("eb = ("+ejbBeanName +")preinvoke(home, " + transAttribute + ");");
	if (isReturnValue) {
	    src.println(JavaType.getName(method.getReturnType()) + " result;");
	}
	src.println("try {");
	src.indentPlus();
	if (isReturnValue) {
	    src.print("result = ");
	}
	src.print("eb." + method.getName() + "(", !isReturnValue);
	for (int p=1; p <= params.length; p++) {
	    src.print("p" + p, false);
	    if (p < params.length) {
		src.print(", ", false);
	    }
	}
	src.println(");", false);
	genCatchBeanException(method, method.getName());
	src.indentMinus();
	src.println("} finally {");
	src.indentPlus();
	src.println("postinvoke(" + transAttribute + ");");
	src.indentMinus();
	src.println("}");
	if (isReturnValue) {
	    src.println("return(result);");
	}
	src.indentMinus(); // end of core
	src.println("}");
    }

    /**
     * Source generation of the 'catch' code following any call of a bean method
     * (such bussiness methods, ejbCreate(), ejbPostCreate(),, .....)
     * The code generated may be:
     *     } catch (RemoteException e) {
     *         ....
     *     } catch (RuntimeException e) {
     *         ....
     *     } catch (Error e) {
     *         ....
     * See the chapter 12.3 in the EJB Specification document
     * @param mInter    the current generated method of the Home of the Remote interface
     * @param mBeanName the name of the associated method of the bean
     */
    void genCatchBeanException(Method mInter, String mBeanName)
	throws GenICException {
	// The "try { ....;" has already been generated
	// The "}" will be generate

	// See if the RemoteException must be catched
	// The RemoteException must be catched if it is in the throws clause
	// of the method of the bean.
	boolean catchRemote = false;
	boolean found = false;
	Method[] methodsBean = ejbBeanClass.getMethods();
	Method mBean = null;
	for (int i=0; ((i<methodsBean.length) && !found); i++) {
	    found = DD.methodsEquals(mInter, methodsBean[i], mBeanName, methodsBean[i].getName());
	    if (found) {
		mBean = methodsBean[i];
	    }
	}
	if (found) {
	    Class[] exceptions = mBean.getExceptionTypes();
	    for (int i=0; ((i<exceptions.length) && (!catchRemote)); i++) {
		catchRemote = ((exceptions[i]).getName()).equals("java.rmi.RemoteException");
	    }
	} else {
	    Trace.outln("GenICRemote() WARNING: Method '" + mBeanName +
			"' associated to the interface method '" +
			mInter + "' not found in the Enterprise Bean");
	}

	src.indentMinus();
	// Catch the System-level exceptions java.rmi.RemoteException
	// See EJB-Specification page 109, &12.3.2
	if(catchRemote){
	    src.println("} catch (RemoteException e) {");
	    src.indentPlus();
	    src.println("Trace.errln(\"Get the following RemoteException thrown by an enterprise Bean's method (rethrow)\");");
	    src.println("Trace.errln(e);");
	    src.println("throw e;");
	    src.indentMinus();
	}
	// Catch the Unchecked exceptions java.lang.RuntimeException and java.lang.Error
	// See EJB-Specification page 109, &12.3.3
	src.println("} catch (RuntimeException e) {");
	src.indentPlus();
	src.println("Trace.errln(\"Get the following RuntimeException thrown by an enterprise Bean's method (convert to a java.rmi.RemoteException)\");");
	src.println("Trace.errln(e);");
	src.println("onUncheckedException(e);");
	src.println("throw new RemoteException(\"Get '\" + e + \"' thrown by an enterprise Bean's method\", e);");
	src.indentMinus();
	src.println("} catch (Error e) {");
	src.indentPlus();
	src.println("Trace.errln(\"Get the following Error thrown by an enterprise Bean's method (convert to a java.rmi.RemoteException)\");");
	src.println("Trace.errln(e);");
	src.println("onUncheckedException(e);");
	src.println("throw new RemoteException(\"Get '\" + e + \"' thrown by an enterprise Bean's method\", e);");

    }

    /**
     * Name of the superclass of the generated class
     * (JBeanSession or JBeanEntity, depending on the Bean type)
     */
    abstract String getSuperClass();

    /**
     * Source generation of the import clauses
     * @exception GenICException in error case
     */
    abstract void   genImport() throws GenICException;

    /**
     * Source generation of the constructor method
     * @exception GenICException in error case
     */
    abstract void   genConstructor() throws GenICException;

    /**
     * Source generation of the getHandle() method
     * (In some case, no source is generated; see sub-classes)
     * @exception GenICException in error case
     */
    abstract void   genGetHandle() throws GenICException;

    /**
     * Source generation of the createData(), removeData(), loadData(), storeData() methods
     * (In some case, no source is generated; see sub-classes)
     * @exception GenICException in error case
     */
    abstract void   genAccessData() throws GenICException;

}

